 // ScanResults.cpp. This program will scan Merlin format output for a genome
// scan and pick out maximum genomewide lod scores for each affection variable
// in the file. These scores are listed as a hypothesis table for input to Pseudo.

// Written by Jan Wigginton
// Last update: 06/17/04

#include "IntArray.h"
#include "Error.h"
#include "Random.h"
#include "Constant.h"
#include "MathMatrix.h"
#include "StringBasics.h"
#include "Parameters.h"
#include "StringArray.h"
#include "MathConstant.h"
#include "StringMap.h"

#include <time.h>
#include <stdio.h>

String scanFileName = "merlin.output";
String hypFileName = "pseudo.hyp";

FILE   * scanFile = NULL;
FILE   * hypFile  = NULL;

Matrix maxChromosomeLods;


int nChromosomes = 0;
int nAffections = 0;
int nLevels = 1;

StringIntMap chromosomeLookup;
StringIntMap affectionLookup;
StringArray  affectionNames;

void ReadParameters(int argc, char *argv[]);
void ReadDimensions();
void ReadData();
void WriteHypothesisTable();

int main(int argc, char *argv[])
   {
   ReadParameters(argc, argv);
   ReadDimensions();

   if (affectionNames.Length() == 0) return 0;

   maxChromosomeLods.Dimension(nAffections, nLevels);
   maxChromosomeLods.Set(_NAN_);

   ReadData();
   WriteHypothesisTable();
   }

void ReadDimensions()
   {
   scanFile = fopen((const char *) scanFileName, "rt");
   if (scanFile == NULL)
      error("ScanResults::Unable to open file %s", (const char *) scanFileName);

   char * flag = NULL;
   int    n_tokens;
   bool skip = true;

   String buffer;
   StringArray tokens;

   String curr_chromosome, curr_affection;
   int chr_index, aff_index;

   while (!feof(scanFile))
      {
      flag = NULL;

      buffer.ReadLine(scanFile);
      tokens.Clear();
      tokens.AddTokens(buffer, WHITESPACE);
      n_tokens = tokens.Length();

      if (n_tokens < 1)
         {
         skip = true;
         continue;
         }

      if (tokens[0] == "Saving")
         {
         skip = true;
         continue;
         }

      if (tokens[0] == "Analysing")
         {
         if (n_tokens < 3)
            error("ScanResults::Unexpected input");

         curr_chromosome = tokens[2];
         if (chromosomeLookup.Find(curr_chromosome) == -1)
            {
            chromosomeLookup.Add(curr_chromosome, nChromosomes);
            nChromosomes++;
            }
         skip = false;
         continue;
         }

      if (tokens[0] == "Phenotype:")
         {
         if (nChromosomes == 0) nChromosomes++;
         if (n_tokens < 4)
            error("Unexpected input");

         curr_affection = tokens[1] + tokens[2];

         if (tokens[3] == "Mean")
            curr_affection = tokens[1] + "[DEVIATES]";

         if (affectionLookup.Find(curr_affection) == -1)
            {
            String new_affection = tokens[1];
            new_affection += " ";
            new_affection += tokens[3] == "Mean" ? " [DEVIATES]" : (const char * ) tokens[2];
            affectionLookup.Add(curr_affection, nAffections);
            affectionNames.Push(new_affection);
            nAffections ++;
            }
         skip = false;
         continue;
         }

      if (skip)  continue;

      if (tokens[0][0] == '=') continue;

      if (tokens[0] == "max" || tokens[0] == "min" || tokens[0] == "Pos")
         continue;


      }

   if (scanFile)
      fclose(scanFile);
   }

void ReadData()
   {
   scanFile = fopen((const char *) scanFileName, "rt");
   if (scanFile == NULL)
      error("ScanResults::Unable to open file %s", (const char *) scanFileName);

   char * flag = NULL;
   int    n_tokens;

   String curr_affection, curr_chromosome;
   int    chr_index = -1, aff_index = -1;

   String buffer;
   StringArray tokens;
   bool skip = true;
   double value;
   int n_columns = 0;

   Vector ** lod_values = new (Vector *)[nAffections];
   for (int aff = 0; aff < nAffections; aff++)
     lod_values[aff] = new Vector[nChromosomes];

   while (!feof(scanFile))
      {
      flag = NULL;
	
      buffer.ReadLine(scanFile);
      tokens.Clear();
      tokens.AddTokens(buffer, WHITESPACE);
      n_tokens = tokens.Length();

      if (n_tokens < 1)
         {
         skip = true;
         continue;
         }

      if (tokens[0] == "Saving")
         {
         skip = true;
         continue;
         }

      if (tokens[0] == "Analysing")
         {
         if (n_tokens < 3)
            error("ScanResults::Unexpected input");

         curr_chromosome = tokens[2];
         chr_index = chromosomeLookup.Find(curr_chromosome);
         skip = false;
         continue;
         }

      if (tokens[0] == "Phenotype:")
         {
         if (n_tokens < 4)
            error("ScanResults::Unexpected input");

         if (nChromosomes == 1) chr_index = 0;

         curr_affection = tokens[1] + tokens[2];

         if (tokens[3] == "Mean")
            curr_affection = tokens[1] + "[DEVIATES]";

         aff_index = affectionLookup.Find(curr_affection);
         skip = false;
         continue;
         }

      if (skip)  continue;

      if (tokens[0][0] == '=') continue;

      if (tokens[0] == "max" || tokens[0] == "min" || tokens[0] == "Pos")
         {
         n_columns = n_tokens;
         continue;
         }

      if (n_tokens != n_columns)
         error("ScanResults::Unexpected input");

      const char * test_value = tokens[n_columns - 2];
      value = strtod(test_value, &flag);
      if (*flag)
         error("In the file %s, the lod value listed (%s) "
                  "is not a valid number\n",
                  (const char *) scanFileName, (const char *) tokens[n_columns - 2]);


      lod_values[aff_index][chr_index].Push(value);
      }

   for (int aff = 0; aff < nAffections; aff++)
      for (int chr = 0; chr < nChromosomes; chr++)
         lod_values[aff][chr].Sort();

   
   IntArray chr_taken;
   int len;
   for (int aff = 0; aff < nAffections; aff++)
      {
      chr_taken.Dimension(nChromosomes);
      chr_taken.Zero();

      double curr_max;
      int    curr_chr;

      for (int lev = 0; lev < nLevels; lev++)
         {
         curr_max = _NAN_;

         for (int chr = 0; chr < nChromosomes; chr++)
            {
            if (chr_taken[chr]) continue;

            len = lod_values[aff][chr].Length();
            if (curr_max == _NAN_ || lod_values[aff][chr][len - 1] > curr_max)
               { 
               curr_max = lod_values[aff][chr][len - 1];
               curr_chr = chr;
               }
            }
         maxChromosomeLods[aff][lev] = curr_max;
         chr_taken[curr_chr] = 1;
         }
      }

   for (int aff = 0; aff < nAffections; aff++)
     if (lod_values[aff])
        delete [] lod_values[aff];

   if (lod_values)
      delete [] lod_values;

   if (scanFile)
      fclose(scanFile);
   }

void WriteHypothesisTable()
   {
   hypFile = fopen((const char *) hypFileName, "wt");
   if (hypFile == NULL)
      error("ScanResults::Unable to open file %s", (const char *) hypFileName);

   for (int aff = 0; aff < nAffections; aff++)
      {
      for (int lev = 0; lev < nLevels; lev++)
         {
         int name_index = affectionLookup.Integer(aff);
         fprintf(hypFile, "Hypothesis_%d  %s\t%5.2lf%5d\n", aff * nLevels + lev,
           (const char *) affectionNames[name_index],
            maxChromosomeLods[aff][lev], lev + 1);
         }
      }

   if (hypFile != NULL)
      fclose(hypFile);
   }

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Analysis Options")
END_LONG_PARAMETERS();

void ReadParameters(int argc, char * argv[])
   {
   ParameterList pl;

   pl.Add(new StringParameter('s', "Merlin output file", scanFileName));
   pl.Add(new StringParameter('h', "Pseudosimulator hypothesis file", hypFileName));
   pl.Add(new IntParameter('l', "Number of levels to scan", nLevels));
   //pl.Add(new LongParameters("Additional Options", longParameters));
   pl.Read(argc, argv);
   pl.Status();
   }


