#include "KAC.h"
#include "Pedigree.h"
#include "MathGenMin.h"
#include "MathStats.h"

#include <math.h>

KAC::KAC(Pedigree & pedigree, StringArray & alias) :
   labels(alias), ped(pedigree)
   {
   affectionCount = familyCount = positionCount = 0;
   min = NULL;
   max = NULL;
   scores = NULL;
   perfamily = NULL;
   }

KAC::~KAC()
   {
   if (min != NULL) delete [] min;
   if (max != NULL) delete [] max;
   if (scores != NULL) delete [] scores;
   }

void KAC::LabelPhenotypes(bool all, bool pairs, bool qtl, bool deviates)
   {
   pheno.Clear();

   if (all)
      for (int a = 0; a < ped.affectionCount; a++)
         pheno.Add(ped.affectionNames[a] + " [ALL]");

   if (pairs)
      for (int a = 0; a < ped.affectionCount; a++)
         pheno.Add(ped.affectionNames[a] + " [Pairs]");

   if (qtl)
      for (int t = 0; t < ped.traitCount; t++)
         pheno.Add(ped.traitNames[t] + " [QTL]");

   if (deviates)
      for (int t = 0; t < ped.traitCount; t++)
         pheno.Add(ped.traitNames[t] + " [QTL, Mean = 0.0]");
   }

void KAC::Prepare(int affections, int families)
   {
   if (min != NULL) delete [] min;
   if (max != NULL) delete [] max;
   if (scores != NULL) delete [] scores;

   affectionCount = affections;
   familyCount = families;
   positionCount = labels.Length();

   if (affections == 0) return;

   min = new Vector[affections];
   max = new Vector[affections];
   scores = new Matrix[affections];

   for (int i = 0; i < affections; i++)
      {
      min[i].Dimension(families);
      min[i].Zero();
      max[i].Dimension(families);
      max[i].Zero();
      scores[i].Dimension(families, positionCount);
      scores[i].Zero();
      }
  }

void KAC::SkipFamily(int serial)
   {
   for (int i = 0; i < affectionCount; i++)
      {
      min[i][serial] = 0.0;
      max[i][serial] = 0.0;
      scores[i][serial].Zero();
      }
   }

void KAC::Set(int affection, int family, double lo, double hi, Vector & zScores)
   {
   min[affection][family] = lo;
   max[affection][family] = hi;
   scores[affection][family] = zScores;
   }

void KAC::SelectChromosome(int chr)
   {
   if (chr == 0)
      chromosome = "";
   else
      chromosome = chr;
      }

void KAC::SelectChromosomeX()
   {
   chromosome = "X ";
   }

double KAC::f(double delta)
   {
   double sum = 0.0;
   for (int i = 0; i < familyCount; i++)
      {
      sum += 2.0 * log(1.0 + delta * scores[currentAffection][i][currentPosition]);
      }
   return -sum;
   }

double KAC::MakeDelta(double value)
   {
   return exp(-exp(value)) * range + minimum;
   }

void KAC::Analyse(Vector & save_results)
   {
   save_results.Clear();
   save_results.Dimension(affectionCount);
   Vector chis, zmeans, deltas;

   for (int affection = 0; affection < affectionCount; affection++)
      {
      AnalyseAffection(affection, chis, deltas, zmeans, false);

      double LOD;
      double max = -99999999999.0;
      for (int pos = 0; pos < positionCount; pos++)
         {
         LOD    = chis[pos] * 0.2171472409516;

         if (deltas[pos] < 0)
            LOD = -LOD;

         if (max < LOD)
            max = LOD;
         }
       save_results[affection] = max;
       }
   }

void KAC::AnalyseAffection(int affection, Vector & chis, Vector & deltas, Vector & zmeans, bool write_results)
   {
   currentAffection = affection;

   chis.Dimension(positionCount);
   chis.Zero();
   deltas.Dimension(positionCount);
   deltas.Zero();
   zmeans.Dimension(positionCount);
   zmeans.Zero();

   double lo = 0.0, hi = 0.0;
   int    informativeFamilies = 0;

   for (int i = 0; i < familyCount; i++)
      if (min[currentAffection][i] != max[currentAffection][i])
         {
         if (min[currentAffection][i] < lo) lo = min[currentAffection][i];
         if (max[currentAffection][i] > hi) hi = max[currentAffection][i];
         informativeFamilies++;
         }
      // else
      //   printf("Family %s is uninformative\n", (const char *) pedi->families[i]->famid);

   if (lo == hi)
      {
      if (write_results)
         printf("No informative affected pairs for %s\n\n",
           (const char *) pheno[currentAffection]);
      return;
      }

   double minDelta = -1.0 / hi;
   double maxDelta = -1.0 / lo;
   double denominator = 1.0 / sqrt(informativeFamilies);

   minimum = minDelta + TINY;
   range   = maxDelta - TINY - minimum;


   // Calculate scores for each analysis position
   for (currentPosition=0; currentPosition<positionCount; currentPosition++)
      {
      double sum = 0.0;

      for (int i = 0; i < familyCount; i++)
         sum += scores[currentAffection][i][currentPosition];

      a = minDelta + TINY; 
      c = maxDelta - TINY;
      b = a + (c - a) * GOLD;
      fb = f(b);
      Brent();

      zmeans[currentPosition] = sum * denominator;
      chis[currentPosition]   = -ScalarMinimizer::fmin;
      deltas[currentPosition] = ScalarMinimizer::min;
      }
   }

void KAC::OutputScores(const char * pos, double zscore, double delta, double chisq)
   {
   double pvalue = ndist(zscore);
   double LOD    = chisq * 0.2171472409516;
   double pdelta = chisq > 0 ? chidist(chisq, 1.0) * 0.5 : 0.5;

   if (delta < 0)
      {
      pdelta = 1.0 - pdelta;
      LOD = -LOD;
      }

   int digits  = pvalue < 1.5e-4 ? 5 : int(log(pvalue*0.06666666)*-0.4343);
   int digits2 = pdelta < 1.5e-4 ? 5 : int(log(pdelta*0.06666666)*-0.4343);

   printf("%20s %7.2f %7.*f %8.3f %6.2f %7.*f\n",
          pos, zscore, digits, pvalue, delta, LOD, digits2, pdelta);
   }

void KAC::OpenPerFamilyFiles()
   {
   perfamily = fopen("merlin.lod", "wt");
   if (perfamily == NULL)
      error("Opening file [merlin.lod] for storing NPL scores for each family\n");
   fprintf(perfamily, "%15s %10s %15s %10s %10s %10s %10s %10s %10s %10s\n",
      "CHROMOSOME", "FAMILY", "TRAIT", "LOCATION", "Z-SCORE", "pLOD", "DELTA", "LOD", "MIN", "MAX");
   }

void KAC::ClosePerFamilyFiles()
   {
   fclose(perfamily);
   printf("NPL scores for individual families stored in [merlin.lod]\n");
   }

void KAC::OutputPerFamilyScores(int affection, int position, double delta)
   {
   for (int i = 0; i < familyCount; i++)
      if (min[affection][i] != max[affection][i])
         {
         // Calculate delta maximized within this family
         double delta_i = scores[affection][i][position] < 0 ?
            -1.0 / max[affection][i] : -1.0 / min[affection][i];

         // k = log10(e)
         const double k = 0.4342944819032;

         // Calculate LOD score maximized within this family
         double lod_i = k * log(1.0 + delta_i * scores[affection][i][position]);

         // Calculate this families contribution to overall LOD
         double partial = k * log(1.0 + delta * scores[affection][i][position]);
         
         fprintf(perfamily, "%14s %10s %15s %10s %10f %10f %10f %10f %10f %10f\n",
            (const char *) chromosome,
            (const char *) ped.families[i]->famid,
            (const char *) pheno[affection], (const char *) labels[position],
            scores[affection][i][position], (delta >= 0) ? partial : -partial,
            delta_i, (delta_i >= 0) ? lod_i : -lod_i, min[affection][i], max[affection][i]);
         }
   }


 
 
 
