// PseudoDataInfo.cpp
// Written by Jan Wigginton

#include "PseudoDataInfo.h"
#include "Error.h"
#include "Constant.h"
#include "MemoryEstimator.h"
#include "PseudoSimulator.h"

#define PSEUDO_SCORE_ALL 0
#define PSEUDO_SCORE_PAIRS 1
#define PSEUDO_SCORE_QTL 2
#define PSEUDO_SCORE_DEV 3

int PseudoDataInfo::maxNChr     = 23;
int PseudoDataInfo::gridPoints  = 1;

String PseudoDataInfo::fileFileName = PseudoSimulator::filePrefix +".list";

PseudoDataInfo::PseudoDataInfo()
   {
   gridOn = gridPoints > 1;
   fileTypeKnown = false;
   useLodFiles = false;
   }

PseudoDataInfo::~PseudoDataInfo()
   {
   FreePositionArrays();
   }

void PseudoDataInfo::FreePositionArrays()
   {
   if (markerSets)
      delete [] markerSets;
   markerSets = NULL;

   if (analysisPositions)
      delete [] analysisPositions;
   analysisPositions = NULL;
   }

void PseudoDataInfo::AllocatePositionArrays()
   {
   markerSets = new IntArray[nChromosomes];
   analysisPositions = new StringArray[nChromosomes];

   for (int chr = 0; chr < nChromosomes; chr++)
      {
      markerSets[chr].Dimension(nAllPositions);
      analysisPositions[chr].Dimension(nAllPositions);
      }
   }

void PseudoDataInfo::Load()
   {
   nChromosomes = nFamilies = nAllPositions = nSkippedPositions = 0;

   chromosomeNames.Clear();
   chromosomeNames.Dimension(maxNChr + 1);

   skippedLookup.Clear();
   familyLookup.Clear();
   positionLookup.Clear();
   chromosomeLookup.Clear();
   possibleAffectionLookup.Clear();

   nPools = ReadFileNames();
   useLodFiles = UseLodFiles();
   useLodFiles ? SetupUsingLodFile((const char *) filesToRead[0])
               : SetupUsingZscoreFile((const char *) filesToRead[0]);

   CheckForData();

   chromosomeNames.Dimension(nChromosomes);
   MakeMarkerSets();
   }

void PseudoDataInfo::CheckForData()   
   {
   bool data_missing = false;

   data_missing |= (possibleAffectionLookup.Length() == 0);
  
   data_missing |= familyLookup.Length() == 0;
   data_missing != chromosomeLookup.Length() == 0;
   
   if (data_missing)
      error("PseudoDataInfo::Some data appears to be missing from file %s\n"
            "Pseudo found %d families, %d chromosomes, and %d affections\n\n", (const char *) filesToRead[0],
            familyLookup.Length(), chromosomeLookup.Length(), possibleAffectionLookup.Length());
   }

int PseudoDataInfo::ReadFileNames()
   {
   FILE * file_file = fopen((const char *) fileFileName, "rb");
   if (file_file == NULL)
      error("PseudoDataInfo::Unable to open file %s", (const char *) fileFileName);

   String      buffer;
   StringArray tokens;
   int         n_tokens;

   filesToRead.Clear();
   while (!feof(file_file))
      {
      buffer.ReadLine(file_file);
      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) < 1) continue;

      if (n_tokens != 1)
         error("PseudoSimulator:Expected one line per entry in file with"
               " input file names. However, current line (%s) has more than"
               " one entry", (const char *) buffer);

      filesToRead.Push(tokens[0]);
      }

   if (filesToRead.Length() == 0) 
      error("PseudoDataInfo::List file (%s) for pseudosimulator is empty", (const char *) fileFileName);

   return filesToRead.Length();
   }

bool PseudoDataInfo::UseLodFiles()
   {
   if (fileTypeKnown) return useLodFiles;

   String buffer;
   StringArray tokens;
   int n_tokens;

   FILE * first_file = fopen((const char *) filesToRead[0], "rb");
   if (first_file == NULL)
      error("PseudoSimulator:Couldn't open file %s", (const char *) filesToRead[0]);

   bool use_lod, found_header = false;
   while (!feof(first_file) && found_header == false)
      {
      buffer.ReadLine(first_file);

      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) <= 1) continue;
      if (tokens[0] != "CHROMOSOME") continue;

      found_header = true;
      use_lod = (n_tokens > 2);
      }

   if (!found_header)
      error("Pseudo is unable to determine your input file format\n"
            "Expecting first data line\n"
            "\n CHROMOSOME ##\n"
            "for zscore format, or \n"
            "\n CHROMOSOME FAMILY TRAIT LOCATION Z-SCORE pLOD DELTA LOD MIN MAX\n"
            "for lod file format\n");

   fileTypeKnown = true;
   return use_lod;
   }

void PseudoDataInfo::SetupUsingZscoreFile(const char * first_file_name)
   {
   String       buffer;
   StringArray  tokens;
   int          n_tokens;
   char *       flag;
   const char * test_value;

   FILE * first_file = fopen(first_file_name, "rb");
   if (first_file == NULL)
      error("PseudoSimulator:Couldn't open file %s", first_file_name);

   String curr_chromosome, curr_family, curr_position, curr_affection;
   int n_possible_affections = 0, n_curr_positions = 0;

   while (!feof(first_file))
      {
      flag = NULL;
      buffer.ReadLine(first_file);

      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) <= 1) continue;

      if (tokens[0] == "CHROMOSOME")
         {
         curr_chromosome = tokens[1];
         if (chromosomeLookup.Find(curr_chromosome) == -1)
            {
            if (curr_chromosome != "X" && curr_chromosome != "x")
               {
               test_value = curr_chromosome;
               int value = strtol(test_value, &flag, 10);
               if (*flag || value < 1 || value > maxNChr)
                  error("In the file %s the chromosome value listed (%s) is not valid",
                         first_file, (const char *) curr_chromosome);
               }

            chromosomeLookup.Add(curr_chromosome, nChromosomes);
            chromosomeNames[nChromosomes++] = curr_chromosome;
            }
         }

      else if (tokens[0] == "POSITIONS")
         {
         n_curr_positions = n_tokens - 1;
         for (int i = 1; i < n_tokens; i++)
            {
            curr_position.printf("%s:%s", (const char *) tokens[i], (const char *) curr_chromosome);

            if (positionLookup.Find(curr_position) == -1 && skippedLookup.Find(curr_position) == -1)
               if (!gridOn || (nAllPositions + nSkippedPositions)%gridPoints == 0)
                  positionLookup.Add(curr_position, nAllPositions++);
               else if (gridOn)
                  skippedLookup.Add(curr_position, nSkippedPositions++);
            }
         }

      else if (tokens[0] == "ANALYSIS" && possibleAffectionLookup.Find(tokens[1]) == -1)
         {
         possibleAffectionLookup.Add(tokens[1], n_possible_affections++);
         curr_affection = tokens[1];
         }

      else if (tokens[0] == "FAMILY")
         {
         if (n_tokens != n_curr_positions + 7)
            error("PseudoSimulator::Number of positions listed for affection %s\n"
                 "on chromosome %s doesn't match the number of z-scores listed for\n"
                 "family %s in file %s", (const char *) curr_affection,
                 (const char *) curr_chromosome, (const char *) curr_family, first_file_name);

         curr_family = tokens[1];
         if (familyLookup.Find(tokens[1]) == -1)
            familyLookup.Add(tokens[1], nFamilies++);
         }
      }

 //  for (int i = 0; i < nChromosomes; i++)
 //    printf("%d %s  %d  %s\n", i, (const char *) chromosomeLookup[i], chromosomeLookup.Integer(i), (const char *) chromosomeNames[i]);
   fclose(first_file);
   }

void PseudoDataInfo::SetupUsingLodFile(const char * first_file_name)
   {
   String      buffer;
   StringArray tokens;
   int         n_tokens;
   char *      flag;

   FILE * first_file = fopen(first_file_name, "rb");
   if (first_file == NULL)
      error("PseudoSimulator:Couldn't open file %s", first_file_name);

   String chromosome, position, last_position;
   String affection_name, last_affection_name, working_affection_name;
   int    n_possible_affections = 0;

   while (!feof(first_file))
      {
      flag = NULL;
      buffer.ReadLine(first_file);

      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) <= 1) continue;

      if (tokens[0] == "CHROMOSOME") continue;

      if (n_tokens != 11)
         error("Input file %s contained unexpected information at line %s",
          first_file_name, (const char *) buffer);

      if (tokens[0] != "X" && tokens[0] != "x")
         {
         const char * test_value = tokens[0];
         int value = strtol(test_value, &flag, 10);
         if (*flag || value < 1 || value > maxNChr)
             error("In the file %s the chromosome value listed (%s) is not valid",
                   first_file_name, (const char *) tokens[0]);
         }

      chromosome = tokens[0];
      if (chromosomeLookup.Find(chromosome) == -1)
         {
         chromosomeNames[nChromosomes] = chromosome;
         chromosomeLookup.Add(chromosome, nChromosomes++);
         }

      affection_name = tokens[2] + ' ' + tokens[3];
      working_affection_name = tokens[2] + tokens[3];
      if (affection_name != last_affection_name && possibleAffectionLookup.Find(affection_name) == -1)
         {
         possibleAffectionLookup.Add((const char *) working_affection_name, n_possible_affections++);
         last_affection_name = affection_name;
         }

      position.printf("%s:%s", (const char *) tokens[4], (const char *) chromosome);
      if (position != last_position)
         {
         if (positionLookup.Find(position) == -1 && skippedLookup.Find(position) == -1)
            {
            if (!gridOn || (nAllPositions + nSkippedPositions)%gridPoints == 0)
               positionLookup.Add(position, nAllPositions++);
            else if (gridOn)
               skippedLookup.Add(position, nSkippedPositions++);
            }
         }
      last_position = position;

      if (familyLookup.Find(tokens[1]) == -1)
         familyLookup.Add(tokens[1], nFamilies++);
      }

   fclose(first_file);
   }

void PseudoDataInfo::MakeMarkerSets()
   {
   AllocatePositionArrays();

   IntArray n_markers(nChromosomes);
   n_markers.Zero();

   int chr_index, div;
   String chromosome, name, position;

   for (int pos = 0; pos < nAllPositions; pos++)
      {
      name = positionLookup[pos];
      div = name.FindChar(':');

      chromosome = name.SubStr(div+1);
      position = name.SubStr(0, div);

      chr_index = chromosomeLookup.Integer(chromosomeLookup.Find(chromosome));
      markerSets[chr_index][n_markers[chr_index]] = pos;
      analysisPositions[chr_index][n_markers[chr_index]++] = name;
      }

   for (int chr = 0; chr < nChromosomes; chr++)
      {
      markerSets[chr].Dimension(n_markers[chr]);
      analysisPositions[chr].Dimension(n_markers[chr]);
      }
   }

int PseudoDataInfo::GetScoreType(const char * score_label)
   {
   String compare_label = score_label;

   if (compare_label.Length() < 3)
      error("PseudoSimulator: Error while reading model parameters.\n"
            "Expecting a comma separated list of outcome names in format\n"
            "name [score_type]. However, current outcome name (%s) doesn't\n"
            "fit this format",
             (const char *) score_label);

   int score_type;
   compare_label.ToUpper();

   if (compare_label == "ALL")
      score_type = PSEUDO_SCORE_ALL;
   else if (compare_label == "PAIRS")
      score_type = PSEUDO_SCORE_PAIRS;
   else if (compare_label == "DEVIATES")
      score_type = PSEUDO_SCORE_DEV;
   else if (compare_label == "QTL")
      score_type = PSEUDO_SCORE_QTL;
   else
      error("PseudoDataInfo: Unknown score type (%s) specified", score_label);

   return score_type;
   }


void PseudoDataInfo::SelectAffectionName(const char * affection_name, int score_type)
   {
   currAffectionName = affection_name;
   workingAffectionName = affection_name;
   String prefix = "";

   switch (score_type)
         {
         case PSEUDO_SCORE_ALL:  prefix = "[ALL]"; break;
         case PSEUDO_SCORE_PAIRS:  prefix = "[Pairs]"; break;
         case PSEUDO_SCORE_QTL:  prefix = "[QTL]"; break;
         case PSEUDO_SCORE_DEV:  prefix = "[QTL, Mean = 0.0]"; break;
         default:  error("PseudoSimulator::Unknown score type chosen");
         }
   currAffectionName += (" " + prefix);

   if (!useLodFiles)
      switch (score_type)
         {
         case PSEUDO_SCORE_ALL:  prefix = ":all"; break;
         case PSEUDO_SCORE_PAIRS:  prefix = ":pairs"; break;
         case PSEUDO_SCORE_QTL:  prefix = ":qtl"; break;
         case PSEUDO_SCORE_DEV:  prefix = ":deviates"; break;
         default:  error("PseudoSimulator::Unknown score type chosen");
         }

   workingAffectionName += prefix;
   }

double PseudoDataInfo::EstimateArrayMemory()
   {
   MemoryEstimator est;

   // markerSets, analysisPositions
   double mem_needed = nChromosomes * est.MemoryForIntArray(nAllPositions);
   mem_needed += nChromosomes * est.MemoryForStringArray(nAllPositions);

   // chromosomesNames, chromosomeLookup
   mem_needed += est.MemoryForStringArray(nChromosomes) + est.MemoryForStringMap(nChromosomes);

   // positionLookup, skippedLookup
   mem_needed += est.MemoryForStringMap(nAllPositions) + est.MemoryForStringMap(nSkippedPositions);

   // familyLookup
   mem_needed += est.MemoryForStringMap(nFamilies);

   // affectionLookup, possibleAffectionLookup
   mem_needed += est.MemoryForStringMap(possibleAffectionLookup.Length()) + est.MemoryForStringMap(nAffections);

   //filesToRead
   mem_needed += est.MemoryForStringArray(nPools);

   return mem_needed;
   }


 
 
 
