// PseudoHypotheses.cpp
// Written by Jan Wigginton

#include "Error.h"
#include "PseudoHypotheses.h"
#include "PseudoDataInfo.h"
#include "Constant.h"
#include "MemoryEstimator.h"
#include "PseudoSimulator.h"

// REVIEW: Deal with RecordNextAffection
// REVIEW: Get nAffections recorded properly

String PseudoHypotheses::hypFileName = PseudoSimulator::filePrefix + ".hyp";

PseudoHypotheses::PseudoHypotheses(PseudoDataInfo & i) :
 nHypotheses(i.nHypotheses), nAffections(i.nAffections), info(i)
   {
   affectionGroups = (IntArray * ) NULL;
   }

PseudoHypotheses::~PseudoHypotheses()
   {
   FreeAffectionGroups();
   }

void PseudoHypotheses::ReadHypotheses()
   {
   FILE * hyp_file = fopen((const char *) hypFileName, "rb");
   if (hyp_file == NULL)
      error("PseudoSimulator:Couldn't open file %s", (const char *) hypFileName);

   String buffer;
   StringArray tokens;
   int n_tokens;
   char * flag = NULL;

   String     hypothesis_label, score_label;
   StringMap  affections;

   nHypotheses = 0;
   nAffections = 0;
   descriptions.Clear();

   while (!feof(hyp_file))
      {
      buffer.ReadLine(hyp_file);

      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) == 0) continue;

      if ((tokens[0].AsUpper() == "TEST") &&
          (tokens[1].AsUpper() == "THRESHOLD") &&
          (tokens[2].AsUpper() == "HITS")&&
          (n_tokens < 4 || tokens[3].AsUpper() == "LABEL" ))
          continue;

      const char * test_value = tokens[n_tokens - 1];
      int rank_to_test = strtol(test_value, &flag, 10);
      if (*flag)
          error("In the file %s, the rank value to test (%s) is not a valid number\n",
                (const char *) hypFileName, (const char *) tokens[n_tokens = 1]);

      ranksToTest.Push(rank_to_test);

      test_value = tokens[n_tokens - 2];
      double fam_lod = strtod(test_value, &flag);
      if (*flag)
            error("In the file %s, the lod value to test (%s) is not a valid number\n",
              (const char *) hypFileName, (const char *) tokens[n_tokens - 2]);

      lodsToTest.Push(fam_lod);

      RecordHypothesis(tokens, affections);
  
      if (labels.Find(tokens[0]) != -1) 
         error("PseudoHypotheses: Duplicate entries for hypothesis %s were found in hypothesis file %s\n", 
                (const char *) tokens[0], (const char *) hypFileName);

      labels.Push(tokens[0]);
      nHypotheses ++;
      }

   if (nHypotheses == 0) 
      error("PseudoHypotheses: Hypothesis file (%s) was empty or didn't contain any valid hypotheses", (const char *) hypFileName);

   AllocateAffectionGroups(affections);
   RecordHypothesisGroups();
   }

void PseudoHypotheses::RecordHypothesis(StringArray & tokens, StringMap & affections)
   {
   descriptions.Push("");

   int curr_token = tokens.Length() - 3;

   if (tokens[curr_token] == "*")
      RecordShortHandHypothesis(affections);
   else
      RecordLongHandHypothesis(tokens, affections);
   }

void PseudoHypotheses::RecordShortHandHypothesis(StringMap & affections)
   {
   descriptions[nHypotheses] = "ALL OUTCOMES";

   int n_possible_affections = info.possibleAffectionLookup.Length();

   for (int aff = 0; aff < n_possible_affections; aff++)
      RecordAffection((const char *) info.possibleAffectionLookup[aff], affections);
   }

#define READ_TOKENS " \t\n\r\f[]"

// Review: Deal with score_label
void PseudoHypotheses::RecordLongHandHypothesis(StringArray & tokens, StringMap & affections)
   {
   String score_label, aff_label, temp_line;
   bool have_score = false;

   int curr_token = 1;
   while (curr_token <= tokens.Length() - 3)
      temp_line += tokens[curr_token++];

   StringArray new_tokens;
   curr_token = new_tokens.AddTokens(temp_line, READ_TOKENS) - 1;

   while (curr_token >= 0)
      {
      new_tokens[curr_token].ExcludeCharacters(",][");
      if (new_tokens[curr_token].Length() == 0)
         {
         curr_token --;
         continue;
         }

      if (!have_score)
        {
        score_label = new_tokens[curr_token];
        have_score = true;
        curr_token--;
        continue;
        }

      aff_label = new_tokens[curr_token];
      int score_type = info.GetScoreType((const char *) score_label);

      UpdateHypothesisDescriptions((const char *) aff_label, (const char *) score_label.AsUpper());
      RecordNextAffection((const char *) aff_label, score_type, affections);

      have_score = false;
      curr_token --;
      score_label = aff_label = "";
      }
   }

void PseudoHypotheses::RecordNextAffection(const char * affection_label,
 int score_type, StringMap & affections)
   {
   info.SelectAffectionName(affection_label, score_type);

   if (info.possibleAffectionLookup.Find(info.workingAffectionName) == -1)
       error("PseudoSimulator is unable to to interpret a hypothesis listed in your\n"
       "hypothesis file.  The current affection name \n"
       "\n\n\t %s \n\n"
       "either isn't present in this data set or is not being interpreted correctly.\n"
       "Hypotheses should be written in the format\n\n"
       "   HYPOTHESIS_LABEL  AFFECTION_1 [SCORE_1], AFFECTION_2 [SCORE_2]  LOD  RANK\n\n"
       "For example\n\n"
       "   HYP1 Affection1 [ALL], Affection2 [PAIRS], Affection3 [ALL]  3.0  1\n"
       "   HYP2 Affection1 [QTL] 2.5 2\n\n"
       "are both valid hypotheses\n",
       (const char *) info.currAffectionName);

   RecordAffection(info.workingAffectionName, affections);
   }

void PseudoHypotheses::RecordAffection(const char * working_affection_name,
 StringMap & affections)
   {
   IntArray * curr_array;
   int        spot;

   if ((spot = affections.Find(working_affection_name))== -1)
      {
      curr_array = new IntArray();
      affections.Add(working_affection_name, (void *) curr_array);
      info.affectionLookup.Add(working_affection_name, info.nAffections++);
      }
   else
      curr_array = (IntArray *) affections.Object(spot);

   curr_array->Push(nHypotheses);
   }

void PseudoHypotheses::UpdateHypothesisDescriptions(const char * aff, const char * score)
   {
   String temp_label;
   temp_label.printf("%s [%s]", aff, score);

   if (descriptions[nHypotheses].Length() > 2)
      descriptions[nHypotheses] += ", ";

   descriptions[nHypotheses] += temp_label;
   }

String PseudoHypotheses::GetDescriptiveString(int hyp, int indent, int line_len)
   {
   StringArray tokens;
   int n_tokens = tokens.AddTokens(descriptions[hyp], ",");

   String spacer(' ', indent);
   String description = tokens[0];

   int line_ct = 0;

   for (int i = 1; i < n_tokens; i++)
      {
      description += ",";
      if (description.Length() - line_ct * line_len + indent > line_len)
         {
         description += "\n" + spacer;
         line_ct ++;
         }
      description += tokens[i];
      }
   return description;
   }

void PseudoHypotheses::AllocateAffectionGroups(const StringMap & affections)
   {
   affectionGroups = new IntArray[nHypotheses];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      affectionGroups[hyp].Dimension(info.nAffections);
      affectionGroups[hyp].Zero();
      }

   for (int aff = 0; aff < info.nAffections; aff++)
      {
      IntArray * curr_array = (IntArray *) affections.Object(aff);

      for (int i = 0; i < curr_array->Length(); i++)
         affectionGroups[(*curr_array)[i]][aff] = 1;
      }
   }

void PseudoHypotheses::FreeAffectionGroups()
   {
   if (affectionGroups)
      delete [] affectionGroups;
   }

double PseudoHypotheses::EstimateArrayMemory()
   {
   //labels
   double mem_needed = MemoryEstimator::MemoryForStringArray(nHypotheses);

   // descriptions
   mem_needed += MemoryEstimator::MemoryForStringArray(nHypotheses);

   //lodsToTest ranksToTest
   mem_needed += MemoryEstimator::MemoryForIntArray(nHypotheses);
   mem_needed += MemoryEstimator::MemoryForVector(nHypotheses);

   // affectionGroups
   mem_needed += nHypotheses * MemoryEstimator::MemoryForIntArray(info.nAffections);

   mem_needed += MemoryEstimator::MemoryForIntArray(nHypotheses);

   return mem_needed;
   }

void PseudoHypotheses::RecordHypothesisGroups()
   {
   nGroups = 0;

   StringIntMap group_map;
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      if (group_map.Find(descriptions[hyp]) == -1)
         group_map.Add(descriptions[hyp], nGroups++);

   groupMembership.Dimension(nHypotheses);
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      groupMembership[hyp] = group_map.Find(descriptions[hyp]);

   nGroups = group_map.Length();
   }

void PseudoHypotheses::BuildGroupLabels(int first_line_len, int max_line_len)
   {
   IntArray * groups = new IntArray[nGroups];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      groups[groupMembership[hyp]].Push(hyp);

   groupLabels.Clear();
   groupLabels.Dimension(nGroups);

   String join, curr_line = "";
   int curr_len_limit = first_line_len;

   for (int grp = 0; grp < nGroups; grp++)
      {
      int n_members = groups[grp].Length();

      for (int grp_mem = 0; grp_mem < n_members; grp_mem++)
         {
         int target = groups[grp][grp_mem];
         groupLabels[grp] += labels[target];
         if (grp_mem == n_members -1) break;

         curr_line += labels[groups[grp][grp_mem]];
         join = grp_mem < n_members - 2 ? ", " : " AND ";
         groupLabels[grp] += join;

         if (curr_line.Length() > curr_len_limit - 5)
            {
            groupLabels[grp] += "\n";
            curr_line = "";
            curr_len_limit = max_line_len;
            }
         }
      }

   if (groups)
      delete [] groups;
   }

 
void PseudoHypotheses::PrintHypothesisDescription(FILE * file, int hyp)
   { 
   fprintf(file, "Hypothesis  :  %s\n", (const char *) labels[hyp]);
   fprintf(file, "Affections  :  %s\n", (const char *) GetDescriptiveString(hyp, 14, 50));
   fprintf(file, "Target Lod  :  %.2f\n", lodsToTest[hyp]);
   fprintf(file, "Target Rank :  %d\n",  ranksToTest[hyp]);
   }
 
