#include "PseudoSimulator.h"
#include "MemoryEstimator.h"

#define BYTES_PER_MEGABYTE 1048576.0
#define BYTES_PER_KILOBYTE 1024.0

void PseudoSimulator::CheckMemoryUsage()
   {
   bufferSize = min(bufferSize, nPseudoSimulations);

   int min_buffer_size = evenSampling ? nPools : 1;

   double mem_needed = 0.0;
   while (true)
      {
      mem_needed = EstimateArrayMemory();

      if (mem_needed <= maxMegabytes * BYTES_PER_MEGABYTE)
         break;

      if (bufferSize < 2 * min_buffer_size)
         break;

      bufferSize /= 2;
      if (evenSampling)
         while (bufferSize%nPools)
            bufferSize++;
      printf("Buffer size is %d\n", bufferSize);
      }

   bool use_megabytes = mem_needed > BYTES_PER_MEGABYTE;

   printf("\nEstimated maximum memory required to perform this simulation is %.1f %s\n\n",
      use_megabytes ? (double) mem_needed / BYTES_PER_MEGABYTE : (double) mem_needed / BYTES_PER_KILOBYTE,
      use_megabytes ? "M" : "K");

   if (mem_needed > maxMegabytes * BYTES_PER_MEGABYTE)
      {
      printf("Pseudo has been terminated because estimated memory allocation for this\n"
              "problem exceeds a preset limit of %d megabytes.\n\n"
              "If your system has resources that exceed this limit and you wish\n"
              "to reset the amount of memory that Pseudo will allow\n"
              "itself to utilize, do so using the --megabytes:9999 option.\n\n"
              "Suggestions for decreasing memory usage :\n\n"
              "1.\tMemory usage may be decreased by turning off the\n"
              "  \t--evenSampling option.\n\n"
              "2.\tWhen a dense marker map is utilized, memory usage may be be reduced\n"
              "  \tby using fewer markers per chromosome by setting --thinData option.\n\n"
              "3.\tWhen performing tests for multiple hypotheses, memory usage\n "
              "  \tmay be reduced by limiting the number of hypotheses tested \n"
              "  \tin a single simulation.\n\n"
              "For further details on each of these options, see program\n"
              "documentation at http://www....\n", maxMegabytes);

      exit(EXIT_FAILURE);
      }
   }

void PseudoSimulator::CheckDiskSpaceUsage()
   {
   MemoryEstimator est;
   double space_est;

   if (maxByChromosome)
      {
      space_est = hypInfo->nGroups * nPseudoSimulations * nChromosomes * est.SpaceForString(maxByChrFormat);
      if (space_est > maxDiskSpace * BYTES_PER_MEGABYTE)
         DiskSpaceError("maximum lod score per chromosome for all hypotheses", "--simpeaks", space_est);
      }

   if (maxByGenome)
      {
      space_est = hypInfo->nGroups * nPseudoSimulations * est.SpaceForString(maxByGenomeFormat);
      if (space_est > maxDiskSpace * BYTES_PER_MEGABYTE)
         DiskSpaceError("maximum lod score per genome for all hypotheses", "--peaks", space_est);
      }

   if (weightSummary)
      {
      space_est = nHypotheses * nFamilies * nChromosomes * est.SpaceForString(var->weightsFormat);
      if (space_est > maxDiskSpace * BYTES_PER_MEGABYTE)
         DiskSpaceError("family weights by chromosome and hypothesis", "--weights", space_est);
      }

   if (repSummary)
      {
      space_est = nFamilies * nPools * nChromosomes * nHypotheses * est.SpaceForString(var->repsFormat);
      if (space_est > maxDiskSpace * BYTES_PER_MEGABYTE)
         DiskSpaceError("detail for family variance weight estimates", "--reps", space_est);
      }
   }

void PseudoSimulator::DiskSpaceError(const char * option_description, const char * option_name,
  double estimated_bytes)
   {
   bool use_megabytes = estimated_bytes > BYTES_PER_MEGABYTE;
   double reported_estimate = use_megabytes ? estimated_bytes /BYTES_PER_MEGABYTE
                                            : estimated_bytes / BYTES_PER_KILOBYTE;

   printf("\n\nPseudo has been terminated because estimated disk space to output\n"
         "%s (option %s)\nis %.0f %s which exceeds a preset limit of %d megabytes. \n\n"
         "If your system has resources that exceed this limit and you wish\n"
         "to reset the amount of disk space that Pseudo will allow itself\n"
         "to utilize when writing to any one output file, do so using the\n"
         "--diskSpace:9999 option.\n\n", option_description, option_name,
         reported_estimate, use_megabytes ? "M" : "K", maxDiskSpace);

   exit(EXIT_FAILURE);
   }

double PseudoSimulator::EstimateArrayMemory()
   {
   MemoryEstimator est;

   //significant_counts
   double mem_needed = nHypotheses * est.MemoryForIntArray(sigChrReported + 1);

   //maxChromosomeLods
   mem_needed += hypInfo->nGroups * sizeof(Matrix *);
   mem_needed += hypInfo->nGroups * est.MemoryForMatrix(nChromosomes, bufferSize);

   // posChromosomes
   mem_needed += nHypotheses * sizeof (IntArray *);
   mem_needed += nHypotheses * nAffections * est.MemoryForIntArray(bufferSize);

   // replicatesTaken
   int dim = evenSampling ? bufferSize : 1;
   mem_needed += nChromosomes * (sizeof(IntArray *) + nFamilies * est.MemoryForIntArray(dim));

   // KAC array
   mem_needed += sizeof(PseudoSimulator);
   mem_needed += nChromosomes * sizeof(KAC);

   // Dynamic memory used in KAC
   // kac.min, kac.max
   mem_needed += 2 * nChromosomes * est.MemoryForVector(nAffections);
   //kac.scores
   mem_needed += nAffections * est.MemoryForMatrix(nFamilies, nAllPositions);
   //kac.labels
   mem_needed += nChromosomes * est.MemoryForStringArray(nAffections);

   mem_needed += sizeof(PseudoZScorePool);
   mem_needed += sizeof(PseudoDataInfo);
   mem_needed += sizeof(PseudoHypotheses);
   mem_needed += sizeof(PseudoVariance);

   if (writePDF)
      mem_needed += sizeof(PseudoPDF);

   mem_needed += pools->EstimateArrayMemory();
   mem_needed += hypInfo->EstimateArrayMemory();
   mem_needed += info.EstimateArrayMemory();
   mem_needed += var->EstimateArrayMemory();
   if (writePDF)
      pdf->EstimateArrayMemory();

   mem_needed += 5000000;
   mem_needed *= 1.05;

   return mem_needed;
   }



 
 
 
