// PseudoPDF.cpp // Written by Jan Wigginton

#include "PseudoPDF.h" 
#include "PseudoSimulator.h" 
#include "PDFhistogram.h" 
#include "PDFlinechart.h" 
#include "Error.h" 
#include "MemoryEstimator.h"

#define PSEUDO_PDF_MIN_BIN 0.1 
#define PSEUDO_N_LOG_THRESHHOLDS 35

String PseudoPDF::pdfFileName = PseudoSimulator::filePrefix + ".pdf";

PseudoPDF::PseudoPDF(PseudoHypotheses & h, PseudoDataInfo & i):
   hypInfo(h), info(i),
   nHypotheses(info.nHypotheses),
   nChromosomes(info.nChromosomes)
   {
   chrMaxCts = genomeMaxCts = NULL;
   }

PseudoPDF::~PseudoPDF()
   {
   FreeResultArrays();
   }

double PseudoPDF::lowerBinLimits[PSEUDO_N_PDF_BINS] =
                   { -15.0, -14.9, -14.8, -14.7, -14.6, -14.5, -14.4, -14.3, -14.2, -14.1,
                      -14.0, -13.9, -13.8, -13.7, -13.6, -13.5, -13.4, -13.3, -13.2, -13.1,
                      -13.0, -12.9, -12.8, -12.7, -12.6, -12.5, -12.4, -12.3, -12.2, -12.1,
                      -12.0, -11.9, -11.8, -11.7, -11.6, -11.5, -11.4, -11.3, -11.2, -11.1,
                      -11.0, -10.9, -10.8, -10.7, -10.6, -10.5, -10.4, -10.3, -10.2, -10.1,
                      -10.0, -9.9, -9.8, -9.7, -9.6, -9.5, -9.4, -9.3, -9.2, -9.1,
                       -9.0, -8.9, -8.8, -8.7, -8.6, -8.5, -8.4, -8.3, -8.2, -8.1,
                       -8.0, -7.9, -7.8, -7.7, -7.6, -7.5, -7.4, -7.3, -7.2, -7.1,
                       -7.0, -6.9, -6.8, -6.7, -6.6, -6.5, -6.4, -6.3, -6.2, -6.1,
                       -6.0, -5.9, -5.8, -5.7, -5.6, -5.5, -5.4, -5.3, -5.2, -5.1,
                       -5.0, -4.9, -4.8, -4.7, -4.6, -4.5, -4.4, -4.3, -4.2, -4.1,
                       -4.0, -3.9, -3.8, -3.7, -3.6, -3.5, -3.4, -3.3, -3.2, -3.1,
                       -3.0, -2.9, -2.8, -2.7, -2.6, -2.5, -2.4, -2.3, -2.2, -2.1,
                       -2.0, -1.9, -1.8, -1.7, -1.6, -1.5, -1.4, -1.3, -1.2, -1.1,
                       -1.0, -0.9, -0.8, -0.7, -0.6, -0.5, -0.4, -0.3, -0.2, -0.1,
                        0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9,
                        1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9,
                        2.0, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9,
                        3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9,
                        4.0, 4.1, 4.2, 4.3, 4.4, 4.5, 4.6, 4.7, 4.8, 4.9,
                        5.0, 5.1, 5.2, 5.3, 5.4, 5.5, 5.6, 5.7, 5.8, 5.9,
                        6.0, 6.1, 6.2, 6.3, 6.4, 6.5, 6.6, 6.7, 6.8, 6.9,
                        7.0, 7.1, 7.2, 7.3, 7.4, 7.5, 7.6, 7.7, 7.8, 7.9,
                        8.0, 8.1, 8.2, 8.3, 8.4, 8.5, 8.6, 8.7, 8.8, 8.9,
                        9.0, 9.1, 9.2, 9.3, 9.4, 9.5, 9.6, 9.7, 9.8, 9.9,
                       10.0, 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7, 10.8, 10.9,
                       11.0, 11.1, 11.2, 11.3, 11.4, 11.5, 11.6, 11.7, 11.8, 11.9,
                       12.0, 12.1, 12.2, 12.3, 12.4, 12.5, 12.6, 12.7, 12.8, 12.9,
                       13.0, 13.1, 13.2, 13.3, 13.4, 13.5, 13.6, 13.7, 13.8, 13.9,
                       14.0, 14.1, 14.2, 14.3, 14.4, 14.5, 14.6, 14.7, 14.8, 14.9 };

void PseudoPDF::AllocateResultArrays()
   {
   genomeMaxCts = new Matrix[nHypotheses];
   chrMaxCts = new Matrix[nHypotheses];

   maxLod.Dimension(nHypotheses);
   minLod.Dimension(nHypotheses);
   maxLod.Set(-99999.0);
   minLod.Set(99999.0);

   logThreshHolds.Dimension(PSEUDO_N_LOG_THRESHHOLDS);
   for (int i = 0; i < PSEUDO_N_LOG_THRESHHOLDS; i++)
      logThreshHolds[i] = i * 0.1;

   logThreshHoldCounts.Dimension(nHypotheses, logThreshHolds.Length());
   logThreshHoldCounts.Zero();

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      chrMaxCts[hyp].Dimension(info.nChromosomes + 2, PSEUDO_N_PDF_BINS);
      chrMaxCts[hyp].Zero();

      genomeMaxCts[hyp].Dimension(3, PSEUDO_N_PDF_BINS);
      genomeMaxCts[hyp].Zero();

      for (int i = 0; i < PSEUDO_N_PDF_BINS; i++)
         {
         chrMaxCts[hyp][0][i] = lowerBinLimits[i];
         chrMaxCts[hyp][1][i] = i < PSEUDO_N_PDF_BINS - 1
                                   ? lowerBinLimits[i + 1]
                                   : lowerBinLimits[PSEUDO_N_PDF_BINS - 1] + 0.1;

         genomeMaxCts[hyp][0][i] = lowerBinLimits[i];
         genomeMaxCts[hyp][1][i] = i < PSEUDO_N_PDF_BINS - 1
                                   ? lowerBinLimits[i + 1]
                                   : lowerBinLimits[PSEUDO_N_PDF_BINS - 1] + 0.1;
         }
      }
   }

double PseudoPDF::EstimateArrayMemory()
   {
   MemoryEstimator est;
   // genomeMaxCts
   double mem_needed = nHypotheses * est.MemoryForMatrix(3, PSEUDO_N_PDF_BINS);

   //chrMaxCts
   mem_needed += nHypotheses * est.MemoryForMatrix(info.nChromosomes + 2, PSEUDO_N_PDF_BINS);

   //logThreshHolds
   mem_needed += est.MemoryForVector(PSEUDO_N_LOG_THRESHHOLDS);

   // logThreshHoldCounts
   mem_needed += est.MemoryForMatrix(nHypotheses, PSEUDO_N_LOG_THRESHHOLDS);

   return mem_needed;
   }

void PseudoPDF::FreeResultArrays()
   {
   if (chrMaxCts)
      delete [] chrMaxCts;

   if (genomeMaxCts)
      delete [] genomeMaxCts;
   }

void PseudoPDF::GraphResults(IntArray * significant_counts)
   {
   pdf.page.SetSize(psLetterR);
   pdf.OpenFile((const char *) pdfFileName);

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      DrawSignificanceHistogram(hyp, significant_counts[hyp]);

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      DrawMaxLodDistribution(hyp);

   DrawLogHits();

   printf("Graphical output written to %s\n", (const char *) pdfFileName);
   pdf.CloseFile();

   #ifdef __DEBUG_GRAPH__
      PrintBinCounts();
   #endif

   }

void PseudoPDF::DrawSignificanceHistogram(int hyp, IntArray & significant_counts)
   {
   PDFHistogram chart;

   int n_cat = PseudoSimulator::sigChrReported + 1;
   Matrix counts(2, n_cat);
   for (int i = 0; i < n_cat; i++)
      {
      counts[0][i] = i;
      counts[1][i] = significant_counts[i];
      }

   chart.SetCategoricalDataCounts(counts);
   chart.xAxis.SetMin(-1);
   chart.xAxis.SetMax(PseudoSimulator::sigChrReported + 1);
   chart.SetSeriesColor(0, 1.0, 0.0, 0.0);

   chart.title.printf("Peak summary for %s", (const char *) hypInfo.labels[hyp]);

   chart.yAxis.label = "Pseudosimulations";
   chart.xAxis.label.printf("Chromosomes with LOD > %5.3lf", hypInfo.lodsToTest[hyp]);

   chart.xAxis.useDiscreteValues = true;
   chart.xAxis.SetStep(1.0);
   chart.xAxis.SetStringTickLabels(true);

   String label = "0";
   int upper = PseudoSimulator::sigChrReported;
   for (int i = 0; i < upper; i++)
      {
      chart.xAxis.SetTickLabel(i, (const char *) label);
      label.printf("%d", i + 1);
      }
   label += "+";
   chart.xAxis.SetTickLabel(upper, (const char *) label);

   chart.yAxis.useDiscreteValues = true;
   chart.useLegend = false;

   chart.Draw(pdf);
   }

void PseudoPDF::DrawMaxLodDistribution(int hyp)
   {
   PDFHistogram chart;
   Matrix all_lods;
   double bar_width;

   DetermineBins(all_lods, genomeMaxCts[hyp], 2, bar_width, maxLod[hyp], minLod[hyp]);

   String label;
   label.printf(" entire genome for %s", (const char *) hypInfo.labels[hyp]);
   DrawMaxLodGraph(hyp, (const char *) label , all_lods, bar_width, true);

   for (int chr = 0; chr < info.nChromosomes; chr++)
      {
      label.printf("chromosome %s for %s", (const char *) info.chromosomeNames[chr],
                                           (const char *) hypInfo.labels[hyp]);
      Matrix chr_results;
      DetermineBins(chr_results, chrMaxCts[hyp], chr + 2, bar_width, maxLod[hyp], minLod[hyp]);
      DrawMaxLodGraph(hyp, (const char *) label, chr_results, bar_width);
      }
   }

void PseudoPDF::DetermineBins(Matrix & new_lods, const Matrix & old_lods,
 int row, double & bar_width, double max, double min)
   {
   int bin = 0, lower = 0, upper = 0;

   while (bin < PSEUDO_N_PDF_BINS)
      {
      if (min == _NAN_ && old_lods[row][bin] > 0 || old_lods[1][bin] >= min)
         {
         lower = bin;
         break;
         }
      bin ++;
      }

   bin = PSEUDO_N_PDF_BINS - 1;
   while (bin > lower)
      if (max == _NAN_ && old_lods[row][bin] > 0 || old_lods[0][bin] <= max )
         {
         upper = bin;
         break;
         }
      else
         bin--;

   double max_lod = max == _NAN_ ? old_lods[1][upper] : max;
   double min_lod = min == _NAN_ ? old_lods[0][lower] : min;

   double range = max_lod - min_lod;
   int cat_per_bar = range > 6.0 ? 2 : range > 1.0 ?  2 : 1;
   bar_width = PSEUDO_PDF_MIN_BIN * cat_per_bar;

   while (lower%cat_per_bar && lower > 0) lower--;
   while ((upper - lower)%cat_per_bar && upper < PSEUDO_N_PDF_BINS) upper++;

   double lower_limit = old_lods[0][lower];
   double total = 0.0;
   new_lods.Dimension(3, 0);
   for (int bin = lower; bin <= upper; bin++)
      {
      if (bin > lower && (bin - lower)%cat_per_bar == 0)
         {
         new_lods[0].Push(lower_limit);
         new_lods[1].Push(lower_limit + bar_width);
         new_lods[2].Push(total);

         lower_limit += bar_width;
         total = 0.0;
         }
      total += bin < PSEUDO_N_PDF_BINS ? old_lods[row][bin] : 0.0;
      }

   new_lods.Dimension(3, new_lods[0].Length());
   }

void PseudoPDF::DrawMaxLodGraph(int hyp, const char * label, Matrix & lods, double bar_width, bool use_green)
   {
   PDFHistogram chart;
   chart.Dimension(lods.rows, lods.cols);

   chart.SetBarWidth(bar_width);
   chart.SetDataCounts(lods);
   chart.xAxis.SetStepMin(bar_width);

   chart.title.printf("Maximum lod score for %s", label);
   chart.subTitle.printf("Number of pseudosimulations: %d", PseudoSimulator::nPseudoSimulations);
   chart.yAxis.label = "Number of chromosomes";
   chart.xAxis.label = "Maximum Lod";
   chart.xAxis.maxMax = maxLod[hyp];
   chart.xAxis.minMin = minLod[hyp];

   use_green ? chart.SetSeriesColor(0, 0.0, 1.0, 0.0) : chart.SetSeriesColor(0, 0.0, 0.0, 1.0);

   chart.useLegend = false;
   chart.yAxis.SetMin(0.0);
   chart.yAxis.useDiscreteValues = true;

   chart.Draw(pdf);
   }

void PseudoPDF::DrawLogHits()
   {
   PDFLineChart chart;
   chart.yAxis.label = "Log10(Hits/Genome)";
   chart.yAxis.usePow10Labels = true;
   chart.xAxis.label = "Lod Cutoff";
   chart.useLegend = false;
   chart.subTitle.printf("Number of pseudosimulations: %d\n", PseudoSimulator::nPseudoSimulations);

   Matrix log_hits_all(0, PSEUDO_N_LOG_THRESHHOLDS), log_hits(1, PSEUDO_N_LOG_THRESHHOLDS);
   log_hits.Zero();

   double denom = log10(PseudoSimulator::nPseudoSimulations);
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      log_hits[0].Set(_NAN_);
      for (int thr = 0; thr < PSEUDO_N_LOG_THRESHHOLDS; thr++)
         if (logThreshHoldCounts[hyp][thr] >= 1e-300)
            log_hits[0][thr] = log10(logThreshHoldCounts[hyp][thr]) - denom;

      chart.Dimension(1, logThreshHolds.Length());
      chart.title.printf("Log10 hits per genome for %s", (const char *) hypInfo.labels[hyp]);
      chart.SetDataValues(logThreshHolds, log_hits);
      chart.SetSeriesColor(0, 1.0, 0.0, 0.0);
      chart.Draw(pdf);
      log_hits_all.StackBottom(log_hits);
      }

   chart.Dimension(nHypotheses, logThreshHolds.Length());
   chart.title.printf("Log10 hits per genome for all hypotheses");
   chart.SetDataValues(logThreshHolds, log_hits_all);
   chart.useLegend = true;
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      chart.SetSeriesLabel(hyp, (const char *) hypInfo.labels[hyp]);

   chart.Draw(pdf);
   }

void PseudoPDF::UpdateCounts(Matrix * max_chromosome_lods, int buffer_size)
   {
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      int grp = hypInfo.groupMembership[hyp];
      UpdateBins(max_chromosome_lods[grp], hyp, buffer_size);
      CountHitsByThreshHold(max_chromosome_lods[grp], hyp, buffer_size);
      }
   }

void PseudoPDF::UpdateBins(Matrix & max_chromosome_lods, int hyp, int buffer_size)
   {
   double max_val, lod_val;
   int max_bin, lod_bin;

   for (int i = 0; i < buffer_size; i++)
      {
      max_val = -999.0;
      for (int chr = 0; chr < info.nChromosomes; chr++)
         {
         lod_val = max_chromosome_lods[chr][i];
         max_val = max(max_val, lod_val);

         lod_bin = lod_val > -1.0 ? 140 : lod_val > -3.0 ?  120 : 0;
         while (lod_bin < 299 && lod_val > lowerBinLimits[lod_bin + 1])
            lod_bin ++;

         chrMaxCts[hyp][chr+2][lod_bin] ++;
         }

      max_bin = max_val > -1.0 ? 140 : max_val > 3.0 ? 120 : 0;
      while (max_bin < 299 && max_val > lowerBinLimits[max_bin + 1])
         max_bin ++;
      genomeMaxCts[hyp][2][max_bin] ++;
      }

   maxLod[hyp] = max(maxLod[hyp], max_chromosome_lods.SafeMax());
   minLod[hyp] = min(minLod[hyp], max_chromosome_lods.SafeMin());
   }

void PseudoPDF::CountHitsByThreshHold(Matrix & max_chromosome_lods, int hyp, int buffer_size)
   {
   logThreshHoldCounts[hyp].Dimension(PSEUDO_N_LOG_THRESHHOLDS);
   logThreshHoldCounts[hyp].Zero();

   for (int chr = 0; chr < nChromosomes; chr++)
      for (int sim = 0; sim < buffer_size; sim++)
         for (int thr = 0; thr < PSEUDO_N_LOG_THRESHHOLDS; thr++)
            if (max_chromosome_lods[chr][sim] > logThreshHolds[thr])
               logThreshHoldCounts[hyp][thr]++;
   }


//////////////////////////// Debug code/////////////////////////////////////////////////////

void PrintBinCounts()
   {
   }



 

