// PseudoSimulator.cpp
// Written by Jan Wigginton

#include "PseudoSimulator.h"
#include "Random.h"
#include "Constant.h"
#include "PDFhistogram.h"
#include "PDFlinechart.h"
#include "MemoryEstimator.h"
#include "PseudoVariance.h"
#include "MathStats.h"
#include "PValue.h"

#include <time.h>

#define SMALLEST_COUNTED_CELL 0

bool PseudoSimulator::weightSummary = false;
bool PseudoSimulator::repSummary = false;
bool PseudoSimulator::maxByChromosome  = false;
bool PseudoSimulator::maxByGenome  = false;
bool PseudoSimulator::hitsByGenome = false;
bool PseudoSimulator::quietOutput      = false;
bool PseudoSimulator::evenSampling     = false;
bool PseudoSimulator::randomSampling   = true;
bool PseudoSimulator::writePDF         = false;

String PseudoSimulator::filePrefix = "pseudo";
String PseudoSimulator::hitsFileName = PseudoSimulator::filePrefix + ".hits";
String PseudoSimulator::maxChrFileName = PseudoSimulator::filePrefix + ".max";
String PseudoSimulator::maxGenomeFileName = PseudoSimulator::filePrefix + ".simmax";

int PseudoSimulator::nPseudoSimulations = 10;
int PseudoSimulator::maxMegabytes       = 1000;
int PseudoSimulator::maxDiskSpace       = 100;
int PseudoSimulator::sigChrReported     = 9;

PseudoSimulator::PseudoSimulator() :
   nPools(info.nPools), 
   nAffections(info.nAffections), 
   nFamilies(info.nFamilies),
   nChromosomes(info.nChromosomes),
   nAllPositions (info.nAllPositions),
   nHypotheses (info.nHypotheses)
   {
   info.Load();

   hypInfo = new PseudoHypotheses(info);
   hypInfo->ReadHypotheses();

   pools   = new PseudoZScorePool(info);
   pdf     = NULL;
   var     = new PseudoVariance(*hypInfo, info);

   sigChrReported = max(hypInfo->ranksToTest.Max(), sigChrReported);

   currMarkers = NULL;
   replicatesTaken = NULL;
   significantCounts = NULL;
   maxChromosomeLods = NULL;
   posChromosomes = NULL;

   maxByChrFile = maxByGenomeFile = hitsFile = (FILE *) NULL;

   bufferSize = max(pools->nPools, nPseudoSimulations);
   }

PseudoSimulator::~PseudoSimulator()
   {
   FreeResultArrays();
   }

void PseudoSimulator::AllocateResultArrays()
   {
   significantCounts = new IntArray[nHypotheses];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      significantCounts[hyp].Dimension(sigChrReported + 1);

   replicatesTaken = new IntArray*[nChromosomes];
      for (int chr = 0; chr < nChromosomes; chr++)
         replicatesTaken[chr] = new IntArray[nFamilies];

   maxChromosomeLods = new Matrix[hypInfo->nGroups];
   maxVals.Dimension(nAffections);

   posChromosomes = new IntArray *[nHypotheses];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      posChromosomes[hyp] = new IntArray[nAffections];

   if (writePDF)
      pdf->AllocateResultArrays();

   var->AllocateResultArrays();
   }

void PseudoSimulator::FreeResultArrays()
   {
   if (significantCounts)
     delete [] significantCounts;
   significantCounts = NULL;

   if (replicatesTaken != NULL)
      {
      for (int chr = 0; chr < nChromosomes; chr++)
         if (replicatesTaken[chr])
            delete [] replicatesTaken[chr];

      delete [] replicatesTaken;
      }
   replicatesTaken = NULL;

   if (maxChromosomeLods)
      delete [] maxChromosomeLods;

   if (posChromosomes)
      {
      for (int hyp = 0; hyp < nHypotheses; hyp++)
         if (posChromosomes[hyp])
            delete [] posChromosomes[hyp];
      delete [] posChromosomes;
      }
   posChromosomes = NULL;
   }

void PseudoSimulator::TestHypotheses()
   {
   if (writePDF)
      pdf = new PseudoPDF(*hypInfo, info);

   OpenSummaryFiles();
   PrintSimulationParameters();
   CheckMemoryUsage();
   CheckDiskSpaceUsage();
   AllocateResultArrays();

   pools->LoadData();

   // Re-create original results and summarize
   RunSimulations(nPools, false);
   SummarizeResults("Actual", nPools);

   // Generate pseudosimulations and summarize
   RunSimulations(nPseudoSimulations, true);
   SummarizeResults("Simulated", nPseudoSimulations);

   //Print most, least influential family/chr for each hypothesis
   var->SummarizeWeights();

   if (writePDF)
      pdf->GraphResults(significantCounts);

   if (repSummary)
      var->PrintFamilyRepSummaries();

   if (weightSummary)
      var->PrintFamilyWeights();

   CloseSummaryFiles();
   }

void PseudoSimulator::PrintSimulationParameters()
   {
   printf("\tNumber of families:           %7d\n", nFamilies);
   printf("\tNumber of positions:          %7d\n", nAllPositions);
   printf("\tNumber of chromosomes:        %7d\n", nChromosomes);
   printf("\tNumber of affections:         %7d\n", nAffections);
   printf("\tNumber of hypotheses:         %7d\n", nHypotheses);
   printf("\tNumber of real simulations:   %7d\n", nPools);
   printf("\tNumber of pseudosimulations:  %7d\n\n", nPseudoSimulations);
   fflush(stdout);
   }

void PseudoSimulator::RunSimulations(int n_simulations, bool random_sample)
   {
   currBuffer = 0;
   InitializeCountArrays(n_simulations, random_sample);

   Pedigree dummy_ped;

   KAC ** kac = new KAC*[nChromosomes];
   for (int i = 0; i < nChromosomes; i++)
      {
      kac[i] = new KAC(dummy_ped, info.analysisPositions[i]);
      kac[i]->pheno.Clear();
      for (int aff = 0; aff < nAffections; aff++)
         kac[i]->pheno.Push(info.affectionLookup[aff]);

      kac[i]->SelectChromosome(info.chromosomeNames[i]);
      kac[i]->Prepare(nAffections, nFamilies);
      }

   for (int curr_sim = 0; curr_sim < n_simulations; curr_sim++)
      {
      if (!evenSampling)
          GetCurrentReplicateSelections(curr_sim, random_sample);

      NewSimulation(kac, curr_sim, random_sample);
      CheckForPositiveSimulation(curr_sim, random_sample);

      if ((curr_sim + 1)%bufferSize == 0 || (curr_sim == n_simulations - 1))
         {
         DumpBuffers(curr_sim, n_simulations, random_sample);
         currBuffer ++;
         }
      }

   if (kac)
      {
      for (int i = 0; i < nChromosomes; i++)
         if (kac[i]) delete kac[i];
      delete [] kac;
      kac = NULL;
      }
   }

void PseudoSimulator::DumpBuffers(int curr_sim, int n_simulations, bool random_sample)
   {
   if (random_sample == false) return;

   //      ??if (curr_buffer * buffer_size + i >= n_simulations) continue;
   //REVIEW Correct this for hypothesis groups
   if (writePDF)
      pdf->UpdateCounts(maxChromosomeLods, bufferSize);

   /// REVIEW: temporary change
   PrintOptionalSummaries(currBuffer == 0);
   ReInitializeBufferArrays(min(bufferSize, n_simulations), random_sample);
   }

void PseudoSimulator::InitializeCountArrays(int n_simulations, bool random_sample)
   {
   ReInitializeBufferArrays(min(bufferSize, n_simulations), random_sample);

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      significantCounts[hyp].Zero();

   nAllCombined = nOneCombined = 0;
   }

void PseudoSimulator::ReInitializeBufferArrays(int n_saved_simulations, bool random_sample)
   {
   for (int grp = 0; grp < hypInfo->nGroups; grp++)
      {
      maxChromosomeLods[grp].Dimension(nChromosomes, n_saved_simulations);
      maxChromosomeLods[grp].Set(-99.0);
      }

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      for (int aff = 0; aff < nAffections; aff++)
         {
         posChromosomes[hyp][aff].Dimension(n_saved_simulations);
         posChromosomes[hyp][aff].Zero();
         }

   int n_saved_samples = evenSampling ? n_saved_simulations : 1;
   for (int chr = 0; chr < nChromosomes; chr++)
      for (int fam = 0; fam < nFamilies; fam++)
         replicatesTaken[chr][fam].Dimension(n_saved_samples);

   if (evenSampling)
      GetEvenReplicateSelections(n_saved_simulations, random_sample);
   }

void PseudoSimulator::NewSimulation(KAC ** kac, int curr_sim, bool random_sample)
   {
   int increment = nPseudoSimulations > 100 ? 100 : 10;

   if (!quietOutput && random_sample && curr_sim && curr_sim%increment == 0)
      {
      printf("Starting simulation: %10d\r", curr_sim);
      fflush(stdout);
      }

   for (int chromosome = 0; chromosome < nChromosomes; chromosome++)
      {
      currMarkers = &info.markerSets[chromosome];
      NewChromosome(*kac[chromosome], curr_sim, chromosome, random_sample);
      }
   }

void PseudoSimulator::NewChromosome(KAC & kac, int sim, int chr, bool random_sample)
   {
   for (int aff = 0; aff < nAffections; aff++)
      {
      kac.min[aff].Zero();
      kac.max[aff].Zero();
      kac.scores[aff].Zero();
      }

   SampleFromPools(kac, sim, chr, random_sample);
   kac.Analyse(maxVals);

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      for (int aff = 0; aff < nAffections; aff++)
         {
         if (hypInfo->affectionGroups[hyp][aff] == 0) continue;

         int grp = hypInfo->groupMembership[hyp];
         maxChromosomeLods[grp][chr][sim%bufferSize]
           = max(maxChromosomeLods[grp][chr][sim%bufferSize], maxVals[aff]);
         posChromosomes[hyp][aff][sim%bufferSize] += (maxVals[aff] > hypInfo->lodsToTest[hyp]);
         }
   }

void PseudoSimulator::CheckForPositiveSimulation(int curr_sim, bool random_sample)
   {
   int sim_spot = curr_sim%bufferSize;
   int rep_slot = evenSampling ? sim_spot : 0;
   int rep;

   int n_greater;
   bool sig_hyp = false, all_sig_hyp = true, one_sig_hyp = false;

   int hyp_group;
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      hyp_group = hypInfo->groupMembership[hyp];

      n_greater = 0;
      for (int chr = 0; chr < nChromosomes; chr++)
         if (maxChromosomeLods[hyp_group][chr][sim_spot]  >= hypInfo->lodsToTest[hyp])
            n_greater ++;
      n_greater = min(n_greater, sigChrReported);

      sig_hyp = (n_greater >= hypInfo->ranksToTest[hyp]);
      all_sig_hyp &= sig_hyp;
      one_sig_hyp |= sig_hyp;

      if (random_sample && sig_hyp)
         for (int fam = 0; fam < nFamilies; fam++)
            for (int chr = 0; chr < nChromosomes; chr++)
               {
               rep = replicatesTaken[chr][fam][rep_slot];
               var->repPosCounts[hyp][chr][fam][rep]++;
               }

      significantCounts[hyp][n_greater]++;
      }

   nAllCombined += all_sig_hyp;
   nOneCombined += one_sig_hyp;
   }

int PseudoSimulator::SampleFromPools(KAC & kac, int curr_sim, int chr, bool random_sample)
   {
   int sample_sim;
   int n_curr_positions = currMarkers->Length();

   for (int fam = 0; fam < nFamilies; fam++)
      {
      sample_sim = replicatesTaken[chr][fam][evenSampling ? curr_sim%bufferSize : 0];

      if (random_sample)
         var->repCounts[chr][fam][sample_sim]++;

      for (int aff = 0; aff < nAffections; aff++)
         {
         kac.min[aff][fam] = pools->affectionMins[aff][sample_sim][fam][chr];
         kac.max[aff][fam] = pools->affectionMaxs[aff][sample_sim][fam][chr];

         for (int pos = 0; pos < n_curr_positions; pos++)
            kac.scores[aff][fam][pos] = pools->affectionScores[aff][sample_sim][fam][(*currMarkers)[pos]];
         }
      }
   return sample_sim;
   }

void PseudoSimulator::GetCurrentReplicateSelections(int curr_sim, bool random_sample)
   {
   for (int chr = 0; chr < nChromosomes; chr++)
      for (int fam = 0; fam < nFamilies; fam++)
         if (random_sample)
            replicatesTaken[chr][fam][0] = int(floor(globalRandom.Next() * nPools));
         else
            replicatesTaken[chr][fam][0] = curr_sim;
   }

void PseudoSimulator::GetEvenReplicateSelections(int n_sampled_simulations, bool random_sample)
   {
   int n_blocks = int(floor(n_sampled_simulations/ pools->nPools));
   int remainder = n_sampled_simulations - n_blocks * pools->nPools;
   if (remainder > 0) n_blocks++;

   for (int chr = 0; chr < nChromosomes; chr++)
      for (int fam = 0; fam < nFamilies; fam++)
         {
         if (random_sample)
            for (int block = 0; block < n_blocks; block++)
               GetPermutationBlock(replicatesTaken[chr][fam], block, pools->nPools);
         else
            replicatesTaken[chr][fam].SetSequence();
         }
   }

void PseudoSimulator::GetPermutationBlock(IntArray & reps_sampled, int block, int block_length)
   {
   int target_length = reps_sampled.Length();

   IntArray remaining;
   remaining.Dimension(block_length);
   remaining.SetSequence();

   int pick, target_index;
   for (int dim = block_length; dim > 0; dim--)
      {
      pick = int(floor(globalRandom.Next() * dim));

      target_index = block * block_length + (dim - 1);
      if (target_index < target_length)
         reps_sampled[target_index] = remaining[pick];

      remaining.Delete(pick);
      }
   }

/*
void PseudoSimulator::GetPermutationBlock(IntArray & reps_sampled, int block, int block_length)
   {
   int target_length = reps_sampled.Length();

   IntArray remaining(block_length);
   remaining.SetSequence();

   int pick, target_index;
   for (int dim = block_length; dim > 0; dim--)
      {
      pick = int(floor(globalRandom.Next() * dim));

      target_index = block * block_length + (dim - 1);
      if (target_index < target_length)
         reps_sampled[target_index] = remaining[pick];

      remaining.Swap(dim, pick);
      }
   }
*/
//int digits = pHat > 0 ? max(2, int(-log10(pHat))) + 1 : 1;
//int digits = 5;


void PseudoSimulator::SummarizeResults(const char * label, int n_simulations)
   {
   int n_categories = significantCounts[0].Length();

   double p_hat, test_lod;
   int    test_rank;

   for(int hyp = 0; hyp < nHypotheses; hyp++)
      {
      SummarizeHypothesisResults(hyp, (const char *) label, n_simulations);
      
      test_rank = hypInfo->ranksToTest[hyp];
      test_lod  = hypInfo->lodsToTest[hyp];

      p_hat = 0.0;
      for (int nsig = 0; nsig < n_categories; nsig++)
         if (nsig >= test_rank)
            p_hat += ((double) significantCounts[hyp][nsig]);
      p_hat /= (double) n_simulations;

      PValue p_print(p_hat, 2, int(log10(n_simulations - 1)) + 1, 1e-5);
      printf("\n\nP-value for %d or more lod%s > %4.3f: %s\n", test_rank,
       test_rank > 1 ? "s" : "", test_lod, (const char *) p_print.output);

      if (n_simulations == nPseudoSimulations)
         {
         double se = sqrt(var->GetEstimatedVariance(hyp, p_hat));
         double max_diff = max(1.0 - p_hat , p_hat);

         bool use_bound = max_diff < se;
         se = use_bound ? max_diff : se;

         DoublePrint se_print(se, 3, se < 1e-3);
         printf("Estimated standard deviation for p-value: %s %s\n", use_bound ? ">" : "",
           (const char *) se_print.output);
         }
      }

   if (n_simulations == nPseudoSimulations)
      SummarizeAllJointly();
   printf("\n\n");
   fflush(stdout);
   }

/*bool use_scientific = se < 0.001 && se > 0;
         int digits_2 = 3;

         String format;
*/

//int digits = 5;
//int digits = pHat > 0 ? max(2, int(-log10(pHat))) + 1 : 1;

void PseudoSimulator::SummarizeAllJointly()
   {
   String fill('=', 16);
   printf("\n\n\nJOINT HYPOTHESES\n%s\n", (const char *) fill);

   double p_one_combined = (double) nOneCombined / (double) nPseudoSimulations;
   double p_all_combined = (double) nAllCombined / (double) nPseudoSimulations;

   PValue p_one(p_one_combined, 1, int(log10(nPseudoSimulations - 1)) + 1, 1e-5);
   PValue p_all(p_all_combined, 1, int(log10(nPseudoSimulations - 1)) + 1, 1e-5);

   String temp, description;
   String all_label, one_label;

   int max_rank_digits = int(log10(hypInfo->ranksToTest.Max())) + 1;

   for (int hyp = 0; hyp < hypInfo->nHypotheses; hyp++)
      {
      int rank = hypInfo->ranksToTest[hyp];
      double lod = hypInfo->lodsToTest[hyp];

      description = hypInfo->GetDescriptiveString(hyp, 27 + max_rank_digits, 50);
      temp.printf("(%d LOD%1s > %3.2lf for %s)", rank, rank == 1 ? " " : "s", lod, (const char *) description);
      all_label += temp;
      all_label += (hyp < hypInfo->nHypotheses - 1) ? "\n      AND " : "\n      = ";

      description = hypInfo->GetDescriptiveString(hyp, 26 + max_rank_digits, 50);
      temp.printf("(%d LOD%1s > %3.2lf for %s)", rank, rank == 1 ? " " : "s", lod, (const char *) description);
      one_label += temp;
      one_label += (hyp < hypInfo->nHypotheses - 1) ? "\n      OR " : "\n      = ";
      }

   printf("\n\nProbability that %s %s (Total hits : %d) \n\n", (const char *) all_label, (const char *) p_all.output, nAllCombined);
   printf("\n\nProbability that %s %s (Total hits : %d)\n\n", (const char *) one_label, (const char *) p_one.output, nOneCombined);
    }

void PseudoSimulator::SummarizeHypothesisResults(int hyp, const char * label, int n_simulations)
   {
   int n_categories = sigChrReported + 1;

   String temp = label;

   String fill('=', temp.Length() + 55);
   printf("\n\n\n%s DISTRIBUTION OF SIGNIFICANT CHROMOSOMES PER SIMULATION\n%s\n\n",
          (const char *) temp.ToUpper(), (const char *) fill);

   hypInfo->PrintHypothesisDescription(stdout, hyp);

//   printf("Hypothesis  :  %s\n", (const char *) hypInfo->labels[hyp]);
//   printf("Affections  :  %s\n", (const char *) hypInfo->GetDescriptiveString(hyp, 14, 50));
//   printf("Target Lod  :  %.2f\n", hypInfo->lodsToTest[hyp]);
   printf("Simulations :  %d\n\n", n_simulations);
   printf("%10s%10s%20s%20s\n", "HITS", "SIMS", "PROB", "CUMULATIVE");
   printf("------------------------------------------------------------\n\n");

   Vector cumulative(n_categories), values(n_categories);
   cumulative.Zero();
   values.Zero();

   double prev = 0.0;
   for (int nsig = n_categories - 1; nsig >= 0; nsig--)
      {
      values[nsig] = ((double) significantCounts[hyp][nsig] ) / ((double) n_simulations);
      cumulative[nsig] = values[nsig] + prev;
      prev += values[nsig];
      }

   int digits_1, digits_2;
   for (int nsig = 0; nsig < n_categories; nsig++)
      {
      digits_1 = values[nsig] > 0 ? max(1, int(-log10(values[nsig]))) : 1;
      digits_2 = cumulative[nsig] > 0 ? max(1, int(-log10(cumulative[nsig]))) : 1;
      printf("%10d%10d%10s%10.*f%10s%10.*f\n", nsig, significantCounts[hyp][nsig], " ",
             digits_1 + 1, values[nsig], " ", digits_2 + 1, cumulative[nsig] );
      }
   }

void PseudoSimulator::PrintOptionalSummaries(bool print_header)
   {
   if (maxByChromosome)
      PrintMaxLodByChromosome();

   if (maxByGenome)
      PrintMaxLodBySimulation();

   if (hitsByGenome)
      PrintHitsBySimulation();
   }

void PseudoSimulator::OpenSummaryFiles()
   {
   if (maxByChromosome)
      {
      maxByChrFile = fopen((const char *) maxChrFileName, "wt");
      if (maxByChrFile == NULL)
         error("Unable to open file %s", (const char *) maxChrFileName);
      }

   if (maxByGenome)
      {
      maxByGenomeFile = fopen((const char *) maxGenomeFileName, "wt");
      if (maxByGenomeFile == NULL)
         error("Unable to open file %s", (const char *) maxGenomeFileName);
      }

   if (hitsByGenome)
      {
      hitsFile = fopen((const char *) hitsFileName, "wt");
      if (hitsFile == NULL)
         error("PseudoSimulator::Unable to open output file %s",(const char *) hitsFileName);
      }

   var->OpenSummaryFiles(weightSummary, repSummary);
   }

void PseudoSimulator::CloseSummaryFiles()
   {
   if (maxByChrFile)
      {
      printf("MLS by chromosome and simulation written to %s\n", (const char *) maxChrFileName);
      fclose(maxByChrFile);
      }

   if (maxByGenomeFile)
      {
      printf("MLS by simulation written to %s\n", (const char *) maxGenomeFileName);
      fclose(maxByGenomeFile);
      }

   if (hitsFile)
      {
      printf("Hits summary written to %s\n",(const char *) hitsFileName);
      fclose(hitsFile);
      }

   var->CloseSummaryFiles();
   printf("\n\n");
   }

void PseudoSimulator::PrintMaxLodByChromosome()
   {
   String header_start = "MAXIMUM LOD SCORE PER CHROMOSOME BY SIMULATED GENOME FOR";
   int header_len = 80;
   hypInfo->BuildGroupLabels(header_len - header_start.Length(), header_len);

   String fill('=', 36);
   for (int grp = 0; grp < hypInfo->nGroups; grp++)
      {
      fprintf(maxByChrFile, "\n\n%s %s\n[SIMULATIONS %d to %d]\n",
             (const char *) header_start, (const char *) hypInfo->groupLabels[grp],
             currBuffer * bufferSize + 1, (currBuffer + 1) * bufferSize);

      fprintf(maxByChrFile, "\n\n%12s%12s%12s\n%s\n", "SIMULATION", "CHROMOSOME", "MAXIMUM LOD", (const char *) fill);

      for (int sim = 0; sim < bufferSize; sim++)
         for (int chr = 0; chr < nChromosomes; chr++)
            fprintf(maxByChrFile, "%12d%12s%12.5f\n", sim + currBuffer * bufferSize + 1,
                   (const char *) info.chromosomeNames[chr], maxChromosomeLods[grp][chr][sim]);
      }
   //save to estimate disk space
   maxByChrFormat = fill;
   }

void PseudoSimulator::PrintMaxLodBySimulation()
   {
   String header_start = "MAXIMUM LOD SCORE BY SIMULATED GENOME FOR";
   int header_len = 70;
   hypInfo->BuildGroupLabels(header_len - header_start.Length(), header_len);

   String fill('=', 30);

   for (int grp = 0; grp < hypInfo->nGroups; grp++)
      {
      fprintf(maxByGenomeFile, "\n\n%s %s\n[SIMULATIONS %d to %d]\n",
        (const char *) header_start, (const char *) hypInfo->groupLabels[grp],
        currBuffer * bufferSize + 1, (currBuffer + 1) * bufferSize);

      fprintf(maxByGenomeFile, "\n\n%10s%15s\n%s\n", "SIMULATION", "MAXIMUM LOD", (const char *) fill);

      double maxVal;
      for (int sim = 0; sim < maxChromosomeLods[grp][0].Length(); sim++)
         {
         maxVal = -999999;
         for (int chr = 0; chr < nChromosomes; chr++)
            maxVal = max(maxVal, maxChromosomeLods[grp][chr][sim]);

         fprintf(maxByGenomeFile, "%10d%15.5lf\n",sim + currBuffer * bufferSize + 1, maxVal);
         }
      }
   // save for estimating disk space
   maxByGenomeFormat = fill;
   }

void PseudoSimulator::PrintHitsBySimulation()
   {
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      PrintHypothesisHits(hyp);
      fprintf(hitsFile, "\n\n");
      }
   }

void PseudoSimulator::PrintHypothesisHits(int hyp)
   {
   String fill('=', hypInfo->labels[hyp].Length() + 71);
   fprintf(hitsFile, "NUMBER OF CHROMOSOME HITS BY AFFECTION AMONG SIMULATIONS POSITIIVE FOR %s\n%s\n\n",
                     (const char *) hypInfo->labels[hyp], (const char *) fill);

   int max_name_length = 0;

   IntArray hyp_affections;
   for (int aff = 0; aff < nAffections; aff++)
      {
      if (hypInfo->affectionGroups[hyp][aff] == 0) continue;
      max_name_length = max(info.affectionLookup[aff].Length(), max_name_length);
      hyp_affections.Push(aff);
      }

   int n_hyp_affections = hyp_affections.Length();
   max_name_length = max(10, max_name_length + 4);

   int n_cols = max( 1, 100 / max_name_length);
   int chr_set = 0;
   int n_sets = (int) ceil((double) (n_hyp_affections + 1)/(double) n_cols);

   int pos_hyp_chromosomes;

   while (chr_set < n_sets)
      {
      String temp, line;
      line.printf("%15s", "PSEUDOREP");

      int start = chr_set * n_cols;
      int end = min(start + n_cols, n_hyp_affections + 1);
      for (int aff = start; aff < end; aff++)
         {
         temp.printf("%*s", max_name_length, aff < n_hyp_affections ?
              (const char *) info.affectionLookup[hyp_affections[aff]] :
              (const char *) hypInfo->labels[hyp]);
         line += temp;
         }
      String fill('=', line.Length() + 1);
      fprintf(hitsFile, "%s\n%s\n", (const char *) line, (const char *) fill);

      int grp = hypInfo->groupMembership[hyp];
      for (int sim = 0; sim < bufferSize; sim++)
         {
         pos_hyp_chromosomes = 0;
         for (int chr = 0; chr < nChromosomes; chr++)
            pos_hyp_chromosomes += (maxChromosomeLods[grp][chr][sim] > hypInfo->lodsToTest[hyp]);

         if (pos_hyp_chromosomes < hypInfo->ranksToTest[hyp])
            continue;

         //fprintf(hitsFile, "%d  of %d\n", pos_hyp_chromosomes, hypInfo->ranksToTest[hyp]);
         line.printf("%15d", sim + currBuffer * bufferSize + 1);
         for (int aff = start; aff < end; aff++)
            {
            temp.printf("%*d", max_name_length,
               aff < n_hyp_affections ? posChromosomes[hyp][hyp_affections[aff]][sim]
                                      : pos_hyp_chromosomes);

            line += temp;
            }
         fprintf(hitsFile, "%s\n", (const char *) line);
         }
      chr_set ++;
      fprintf(hitsFile, "\n\n");
      }
   }



 
 
 
