// PseudoSimulator.h:
// Written by:  Jan Wigginton
// Last update: 09/15/05

#ifndef _PSEUDOSIMULATOR_H_
#define _PSEUDOSIMULATOR_H_

#include "KAC.h"
#include "StringArray.h"
#include "MathMatrix.h"
#include "StringBasics.h"
#include "PseudoPDF.h"
#include "StringMap.h"
#include "PseudoZScorePool.h"
#include "PseudoVariance.h"

class PseudoSimulator
   {
   public:

      PseudoSimulator();
      ~PseudoSimulator();

      // Maximum megabytes allowed for memory usage -- simulations requiring more
      // than maxMegabytes will automatically abort
      static int maxMegabytes;

      // Maximum megabytes to be written to any one output file -- simulations requiring
      // more than maxDiskSpace will automatically abort
      static int maxDiskSpace;

      // Indicator for printing list of maximum lod score per chromosome for each
      // pseudoreplicate genome scan (--peaks)
      static bool maxByChromosome;

      // Indicator for printing list of maximum genomewise lod score for each
      // pseudoreplicate genome scan (--simpeaks)
      static bool maxByGenome;

      //Indicator for a breakdown of chromosome hits for each hypothesis by
      //component affection for each pseudoreplicate genome scan (--hits)
      static bool hitsByGenome;

     // Indicator for printing of family weight summaries for each chromosome/family combination
      static bool weightSummary;

      // Indicator for printing of detailed family weight summary for each chromosome/family combination
      static bool repSummary;

      //Indicator for pdf output
      static bool writePDF;

      //Indicator to suppress printing of progress reports for simulation
      static bool quietOutput;

      //Indicators for type of replicate pool sampling -- evenSampling implies
      //z-score pools are sampled evenly
      static bool evenSampling, randomSampling;

      //File name for storing information on hits broken down by individual affection variables
      static String hitsFileName;

      // Name of output file with information on maximum lods by chromosome for each
      // pseudoreplicate genome scan
      static String maxChrFileName;

      //Name of output file with information on genomewide max lods for each
      //pseudoreplicate genome scan
      static String maxGenomeFileName;

      // Prefix to be used for output file names.
      static String filePrefix;

      // Number of pseudoreplicate genome scans to generate
      static int nPseudoSimulations;

      // maximum number of significant chromosomes reported
      static int sigChrReported;

      void TestHypotheses();

   private:
      // Structure containing parameters/searchable indices for data
      PseudoDataInfo     info;

      // Loads and stores information pertaining to hypotheses to be tested
      PseudoHypotheses * hypInfo;

      // Loads and stores data for z-score pools
      PseudoZScorePool * pools;

      //Contains methods and data relating to family variance weights and variance estimation
      PseudoVariance   * var;

      // Used for PDF output
      PseudoPDF        * pdf;

      // Size of simulation blocks completed in memory, current simulation block
      int bufferSize, currBuffer;

      // Max chromosome lods
      FILE * maxByChrFile;

      // Max simulation lods by genome
      FILE * maxByGenomeFile;

      // Hits by affection for each hypothesis and replicate
      FILE * hitsFile;

      // Simulated p-value
      double pHat;

      // Number of simulations with all ("ALL") or at least one ("OR") hypotheses/is true
      int nAllCombined,  nOneCombined;

      // significantCounts[hyp][j] = number of pseudoreplicates with j "+"
      // chromosomes for hypothesis hyp
      IntArray * significantCounts;

      // Replicates sampled by chromosome, family and simulation
      // replicatesTaken[fam][chr][0] =  replicate sampled for family fam
      // on chromosome chr for simulation 0 in pools zscore array
      IntArray ** replicatesTaken;

      // maxChromosomeLods[hyp][chr][sim_spot] = Maximum simulated lod score by
      // hypothesis, chromosome and simulation
      Matrix * maxChromosomeLods;

      // Counts of + chromosomes with respect to each hypothesis by affection and simulation
      IntArray ** posChromosomes;

      // Pointer to current marker set indices
      IntArray * currMarkers;

      // Random seed (default is int(time(NULL)))
      int randomSeed;


      // Variables to streamline code performance, maintenance or readabililty

      // Aliases used to simplify notation (all initialized to entries in info data structure)
      int & nPools, & nAffections, & nFamilies, & nChromosomes, & nAllPositions, & nHypotheses;

      // String format specification for various output files -- useful for maintaining
      // outfile file size checks
      String maxByGenomeFormat, maxByChrFormat, hitsByGenomeFormat;

      //Global vectors used to avoid reallocation in innner loops
      Vector maxVals;

      void   PrintSimulationParameters();

      void GetEvenReplicateSelections(int n_simulations, bool random_sample);
      void GetCurrentReplicateSelections(int curr_sim, bool random_sample);
      void GetPermutationBlock(IntArray & reps_sampled, int block, int block_length);

      void RunSimulations(int n_simulations, bool random_sample);
      void NewSimulation(KAC ** kac, int curr_sim, bool random_sample);
      void NewChromosome(KAC & kac, int curr_sim, int chromosome, bool random_sample);
      int  SampleFromPools(KAC & kac, int sim, int chromosome, bool random_sample);

      void InitializeCountArrays(int n_simulations, bool random_sample);
      void ReInitializeBufferArrays(int n_simulations, bool random_sample);
      void DumpBuffers(int curr_sim, int n_simulations, bool random_sample);
      void CheckForPositiveSimulation(int curr_sim, bool random_sample);

      void SummarizeResults(const char * label, int n_simulations);
      void SummarizeHypothesisResults(int hyp, const char * label, int n_simulations);
      void SummarizeAllJointly();

      void PrintOptionalSummaries(bool print_header);
      void OpenSummaryFiles();
      void CloseSummaryFiles();

      void PrintMaxLodByChromosome();
      void PrintMaxLodBySimulation();

      void PrintHitsBySimulation();
      void PrintHypothesisHits(int hyp);

      void AllocateResultArrays();
      void FreeResultArrays();

      // Routines to ensure that resource limits are enforced
      void   CheckMemoryUsage();
      void   CheckDiskSpaceUsage();
      void   DiskSpaceError(const char * description, const char * name, double bytes);
      double EstimateArrayMemory();
      };

#endif
 
 
 
