// PseudoVariance.cpp
// Written by Jan Wigginton

#include "PseudoVariance.h"
#include "MemoryEstimator.h"
#include "PseudoSimulator.h"
#include "MathConstant.h"

String PseudoVariance::weightsFileName = PseudoSimulator::filePrefix + ".weights";
String PseudoVariance::repsFileName = PseudoSimulator::filePrefix + "reps";

#define PSEUDO_VARIANCE_MIN_CELL 0

PseudoVariance::PseudoVariance(PseudoHypotheses & h, PseudoDataInfo & i):
   info(i),
   hypInfo(h), 
   nPools(i.nPools),
   nFamilies(i.nFamilies), 
   nChromosomes(i.nChromosomes),
   nHypotheses(i.nHypotheses)
   {
   weightsEstimated = false;
   repsFile = weightsFile = NULL;

   AllocateResultArrays();
   }

PseudoVariance::~PseudoVariance()
   {
   FreeResultArrays();
   }

void PseudoVariance::AllocateResultArrays()
   {
   familyWeights = new Matrix[nHypotheses];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      familyWeights[hyp].Dimension(nFamilies, nChromosomes);
      familyWeights[hyp].Zero();
      }

   repCounts = new IntArray *[nChromosomes];
   for (int chr = 0; chr < nChromosomes; chr ++)
      {
      repCounts[chr] = new IntArray[nFamilies];
      for (int fam = 0; fam < nFamilies; fam++)
         {
         repCounts[chr][fam].Dimension(nPools);
         repCounts[chr][fam].Zero();
         }
      }

   repPosCounts = new IntArray **[nHypotheses];
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      repPosCounts[hyp] = new IntArray *[nChromosomes];
      for (int chr = 0; chr < nChromosomes; chr++)
         {
         repPosCounts[hyp][chr] = new IntArray[nFamilies];
         for (int fam = 0; fam < nFamilies; fam++)
            {
            repPosCounts[hyp][chr][fam].Dimension(nPools);
            repPosCounts[hyp][chr][fam].Zero();
            }
         }
      }
   }

void PseudoVariance::FreeResultArrays()
   {
   if (familyWeights)
     delete [] familyWeights;

   if (repCounts != NULL)
      {
      for (int chr = 0; chr < nChromosomes; chr++)
         if (repCounts[chr])
            delete [] repCounts[chr];
      delete [] repCounts;
      }
   repCounts = NULL;

   if (repPosCounts != NULL)
      {
      for (int hyp = 0; hyp < nHypotheses; hyp++)
         {
         if (repPosCounts[hyp] != NULL)
             {
             for (int chr = 0; chr < nChromosomes; chr++)
                if (repPosCounts[hyp][chr])
                   delete [] repPosCounts[hyp][chr];
             delete [] repPosCounts[hyp];
             }
         }
      delete [] repPosCounts;
      }
   repPosCounts = NULL;
   }

double PseudoVariance::EstimateArrayMemory()
   {
   MemoryEstimator est;
   // repCounts
   double mem_needed = nChromosomes * (sizeof(IntArray *) + nFamilies * est.MemoryForIntArray(nPools));

   // repPosCounts
   mem_needed += nHypotheses * (sizeof(IntArray **) + nChromosomes * sizeof(IntArray *));
   mem_needed += nHypotheses * nChromosomes * nFamilies * est.MemoryForIntArray(nPools);

   // familyWeights
   mem_needed += nHypotheses * est.MemoryForMatrix(nFamilies, nChromosomes);

   return mem_needed;
   }

void PseudoVariance::OpenSummaryFiles(bool weights, bool reps)
   {
   if (weights)
      {
      weightsFile = fopen((const char *) weightsFileName, "wt");
      if (weightsFile == NULL)
         error("PseudoVariance::Unable to open file %s",(const char *) weightsFileName);
      }

   if (reps)
      {
      repsFile = fopen((const char *) repsFileName, "wt");
      if (repsFile == NULL)
         error("PseudoVariance::Unable to open file %s", (const char *) repsFileName);
      }
   }

void PseudoVariance::CloseSummaryFiles()
   {
   if (repsFile != NULL)
      {
      printf("Replicate summary written to %s\n",(const char *) repsFileName);
      fclose(repsFile);
      }

   if (weightsFile != NULL)
      {
      printf("Family variance weights written to %s\n", (const char *) weightsFileName);
      fclose(weightsFile);
      }
   }

void PseudoVariance::PrintFamilyRepSummaries()
   {
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      String label = "REPLICATE SUMMARY RESULTS FOR ";
      String fill_2('=', label.Length() + hypInfo.labels[hyp].Length());
      fprintf(repsFile, "%s%s\n%s\n", (const char *) label, (const char *) hypInfo.labels[hyp], (const char *) fill_2);

      for (int chr = 0; chr < nChromosomes; chr++)
         PrintFamilyRepSummary(chr, hyp);

      }
   }
/*
void PseudoVariance::PrintHypothesisHeader(FILE * file, int hyp, const char * summary_label)
   {
   fprintf(file, "%s FOR %s\n\n", summary_label, (const char *) hypInfo.labels[hyp]);
   fprintf(file, "Number of replicates:        %5d\n", info.nPools);
   fprintf(file, "Number of pseudosimulations: %5d\n", PseudoSimulator::nPseudoSimulations);
   fprintf(file, "Rank to be tested:           %5d\n", hypInfo.ranksToTest[hyp]);
   fprintf(file, "Lod to be tested:            %5.2lf\n\n\n", hypInfo.lodsToTest[hyp]);
   }
*/

void PseudoVariance::PrintFamilyRepSummary(int chr, int hyp)
   {
   double ratio;
   String fill_top('-', 52);

   for (int fam = 0; fam < nFamilies; fam++)
      {
      fprintf(repsFile, "\n[CHROMOSOME %s FAMILY %s]\n\n", (const char *) info.chromosomeNames[chr], 
        (const char *) info.familyLookup[fam]); 
      fprintf(repsFile, "\n%13s%13s%13s%13s\n%s\n","REPLICATE", "SELECTIONS",
                         "POSITIVE", "PROP POS", (const char *) fill_top);

      for (int rep = 0; rep < nPools; rep++)
         if (repCounts[chr][fam][rep] > 0)
            {
            ratio = (double) repPosCounts[hyp][chr][fam][rep]/ (double) repCounts[chr][fam][rep];
            fprintf(repsFile, "%13d%13d%13d%13.5lf\n", rep + 1, repCounts[chr][fam][rep],
                    repPosCounts[hyp][chr][fam][rep], ratio);
            }
         else
            fprintf(repsFile, "%13d%13d%13d%13s\n", rep + 1, repCounts[chr][fam][rep],
                    repPosCounts[hyp][chr][fam][rep], "-");

      String fill('-', 7);
      fprintf(repsFile, "%45s%s\n%26s%13s%13.5f\n\n", " " , (const char *) fill, " ", "WEIGHT : ",  
        familyWeights[hyp][fam][chr]);
      }
   repsFormat = fill_top;
   }

void PseudoVariance::PrintFamilyWeights()
   {
   String fill('-', 22);
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      String fill_header('=', hypInfo.labels[hyp].Length() + 46);

      fprintf(weightsFile, "\n\nVARIANCE WEIGHTS BY FAMILY AND CHROMOSOME FOR %s\n%s\n\n",
              (const char *) hypInfo.labels[hyp], (const char *) fill_header);

      for (int chr = 0; chr < nChromosomes; chr++)
         {
         fprintf(weightsFile, "\n[ CHROMOSOME %s ]\n", (const char *) info.chromosomeNames[chr]);
         fprintf(weightsFile, "\n%10s %10s\n%s\n", "FAMILY", "WEIGHT", (const char *) fill);
         for (int fam  = 0; fam < nFamilies; fam++)
            fprintf(weightsFile, "%10s %10.5f\n",
                    (const char *) info.familyLookup[fam], familyWeights[hyp][fam][chr]);
         }
      }

   weightsFormat = fill;
   }

double PseudoVariance::GetEstimatedVariance(int hyp, double p_hat)
   {
   EstimateFamilyWeights();

   double weight_sum = 0.0;
   for (int fam = 0; fam < nFamilies; fam++)
      weight_sum += familyWeights[hyp][fam].Sum();

   double variance = weight_sum - nFamilies * nChromosomes * p_hat;
   variance *= p_hat / (double) nPools;
   variance += p_hat / (double) PseudoSimulator::nPseudoSimulations;

   return max(variance, 0.0);
   }

void PseudoVariance::EstimateFamilyWeights()
   {
   if (weightsEstimated) return;

   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      familyWeights[hyp].Zero();

      int n_pos_tot = (repPosCounts[hyp][0][0]).Sum();
      if (n_pos_tot == 0)
         continue;

      Matrix proportions(nChromosomes, nPools);

      for (int fam = 0; fam < nFamilies; fam++)
         {
         proportions.Zero();

         for (int chr = 0; chr < nChromosomes; chr++)
            for (int rep = 0; rep < nPools; rep++)
               proportions[chr][rep] = repCounts[chr][fam][rep] > 0 ?
               (double) repPosCounts[hyp][chr][fam][rep] / (double) repCounts[chr][fam][rep] : 0.0;

         for (int chr = 0; chr < nChromosomes; chr++)
            {
            for (int rep = 0; rep < nPools; rep++)
               if (repCounts[chr][fam][rep] > PSEUDO_VARIANCE_MIN_CELL )
                  familyWeights[hyp][fam][chr] += proportions[chr][rep] * repPosCounts[hyp][chr][fam][rep];

            familyWeights[hyp][fam][chr] /= (double) n_pos_tot;
            }
         }
      }

   weightsEstimated = true;
   }

void PseudoVariance::SummarizeWeights()
   {
   for (int hyp = 0; hyp < nHypotheses; hyp++)
      {
      SummarizeWeightsByChromosome(hyp);
      SummarizeWeightsByFamily(hyp);
      }
   }

void PseudoVariance::SummarizeWeightsByChromosome(int hyp)
   {
   String label = "MAXIMUM AND MINIMUM WEIGHTS BY CHROMOSOME";
   String fill('=', label.Length());
   printf("\n\n%s\n%s\n\n", (const char *) label, (const char *) fill);

   hypInfo.PrintHypothesisDescription(stdout, hyp);

   String fill_2('-', 73);
   printf("\n%7s |  %10s  |  %10s %10s  |  %10s %10s \n%s\n", "CHR", "AVERAGE", "MAX WT", "FAMILY", "MIN WT", "FAMILY", 
       (const char *) fill_2);

   Vector   max_weights(nChromosomes), min_weights(nChromosomes), averages(nChromosomes);
   IntArray max_families(nChromosomes), min_families(nChromosomes);

   GetExtremeFamilies(hyp, true, max_weights, max_families, averages);
   GetExtremeFamilies(hyp, false, min_weights, min_families, averages);

   String max_line, min_line;
   for (int chr = 0; chr < nChromosomes; chr++)
      {
      min_line.printf("%10.5f %10s", min_weights[chr], (const char *) info.familyLookup[min_families[chr]]);
         
      max_line.printf("%10.5f %10s", max_weights[chr], (const char *) info.familyLookup[max_families[chr]]);

      printf("%7s |  %10.5f  |  %s  |  %s\n", (const char *) info.chromosomeNames[chr], averages[chr], 
               (const char *) max_line, (const char *) min_line);
      }

   double max_overall = familyWeights[hyp].SafeMax();
   double min_overall = familyWeights[hyp].SafeMin(); 
   double avg_overall = familyWeights[hyp].SafeMean();

   printf("%s\n%7s |  %10.5f  |  %10.5f %10s  |  %10.5f %10s\n\n\n", (const char *) fill_2, "OVERALL", avg_overall, max_overall, " ", min_overall, " ");
   }

void PseudoVariance::GetExtremeFamilies(int hyp, bool get_max, Vector & weights, IntArray & families, Vector & averages)
   {
   int max_dir = get_max ? 1 : -1;
   //int max_fam = -1, max_chr = -1;

   double max_weight = max_dir * -1 * 9999999.0;
   weights.Set(max_dir * -1.0 * 99999999.0);
   families.Set(-1);
   averages.Zero();

   double total; 

   for (int chr = 0; chr < nChromosomes; chr++)
      {
      total = 0.0;
      for (int fam = 0; fam < nFamilies; fam++)
         {
         total += familyWeights[hyp][fam][chr];
         if (max_dir * familyWeights[hyp][fam][chr] > max_dir * weights[chr])
            {
            weights[chr] = familyWeights[hyp][fam][chr];
            families[chr] = fam;
            }
         }
      averages[chr] = total / (double) max(nFamilies, 1);
      }  
   }
 
void PseudoVariance::SummarizeWeightsByFamily(int hyp)
   {
   String label = "MAXIMUM AND MINIMUM WEIGHTS BY FAMILY";
   String fill('=', label.Length());
   printf("\n\n%s\n%s\n\n", (const char *) label, (const char *) fill);
      
   hypInfo.PrintHypothesisDescription(stdout, hyp);
      
   String fill_2('-', 76);
   printf("\n%10s |  %10s  |  %10s %10s  |  %10s %10s\n%s\n", "FAMILY", "AVERAGE", "MAX WT", "CHR",  "MIN WT", "CHR",
          (const char *) fill_2);
       
   Vector   max_weights(nFamilies), min_weights(nFamilies), averages(nFamilies);
   IntArray max_chromosomes(nFamilies), min_chromosomes(nFamilies);
   String   max_line, min_line;
      
   GetExtremeChromosomes(hyp, true,  max_weights, max_chromosomes, averages);
   GetExtremeChromosomes(hyp, false, min_weights, min_chromosomes, averages);
      
   for (int fam = 0; fam < nFamilies; fam++)
      {
      min_line.printf("%10.5f %10s", min_weights[fam], (const char *) info.chromosomeNames[min_chromosomes[fam]]);
         
      max_line.printf("%10.5f %10s", max_weights[fam], (const char *) info.chromosomeNames[max_chromosomes[fam]]);
         
      printf("%10s |  %10.5f  |  %s  |  %s \n", (const char *) info.familyLookup[fam], averages[fam], (const char *) max_line,
                (const char *) min_line);
      }

   double max_overall = familyWeights[hyp].SafeMax();
   double min_overall = familyWeights[hyp].SafeMin();
   double avg_overall = familyWeights[hyp].SafeMean();

   printf("%s\n%10s |  %10.5f  |  %10.5f %10s  |  %10.5f %10s \n\n\n", (const char *) fill_2, "OVERALL",  avg_overall, max_overall, " " , min_overall, " ");  
   }

void PseudoVariance::GetExtremeChromosomes(int hyp, bool get_max, Vector & weights, IntArray & chromosomes, Vector & averages)
   {
   int max_dir = get_max ? 1 : -1;

   weights.Set(max_dir * -1.0 * 9999999999.0);
   chromosomes.Set(-1);
   averages.Zero();

   double total;
   for (int fam = 0; fam < nFamilies; fam++)
      {
      total = 0.0;
      for (int chr = 0; chr < nChromosomes; chr++)
         {
         total += familyWeights[hyp][fam][chr];
         if (max_dir * familyWeights[hyp][fam][chr] > max_dir * weights[fam])
            {
            weights[fam] = familyWeights[hyp][fam][chr];
            chromosomes[fam] = chr;
            }
         }
      averages[fam] = total / (double) nChromosomes;
      }
   }
 
 
