// PseudoZScorePool.cpp
// Written by Jan Wigginton

#include "PseudoZScorePool.h"
#include "MemoryEstimator.h"
#include "PseudoDataInfo.h"
#include "Error.h"
#include "Constant.h"


PseudoZScorePool::PseudoZScorePool(PseudoDataInfo & i) : info(i), nPools(i.nPools)
   {
   affectionScores = affectionMins = affectionMaxs = NULL;
   }

PseudoZScorePool::~PseudoZScorePool()
   {
   FreeInputArrays();
   }

void PseudoZScorePool::LoadData()
   {
   AllocateInputArrays();
   info.UseLodFiles() ? ReadRawLodData() : ReadRawZScoreData();
   printf("\n\n");
   }

void PseudoZScorePool::ReadRawLodData()
   {
   int n_files = info.filesToRead.Length();
   for (int sim = 0; sim < n_files; sim++)
      {
      printf("Reading information from file %s\n", (const char *) info.filesToRead[sim]);
      ReadRawLodFile(sim);
      }
   }

void PseudoZScorePool::ReadRawLodFile(int sim)
   {
   FILE * curr_file = fopen((const char *) info.filesToRead[sim], "rb");
   if (!curr_file)
      error("PseudoSimulator::Couldn't find simulation result file %s",
               (const char *) info.filesToRead[sim]);

   String       buffer;
   StringArray  tokens;
   char *       flag;
   const char * test_value;
   int          n_tokens;

   String curr_family_name, last_family_name;
   String affection_name, last_affection_name, working_affection_name;
   String chromosome, last_chromosome;
   String curr_position, last_position;

   int fam_index = -1, chr_index = -1, pos_index = -1, aff_index = 0;
   double min, max, zscore;

   while (!feof(curr_file))
      {
      flag = NULL;

      buffer.ReadLine(curr_file);
      tokens.Clear();
      if ((n_tokens = tokens.AddTokens(buffer, WHITESPACE)) < 1) continue;

      if (tokens[0] == "CHROMOSOME") continue;

      if (n_tokens != 11)
         error("PseudoSimulator::Input line should have 11 columns, "
               "but current line\n%s\nfrom file %s does not",
               (const char *) buffer, (const char *) info.filesToRead[sim]);

      affection_name = tokens[2] + tokens[3];
      aff_index = info.affectionLookup.Find(affection_name);
      if (aff_index == -1)
         continue;

      chromosome = tokens[0];
      if (chromosome != last_chromosome)
         {
         if ((chr_index = info.chromosomeLookup.Find(chromosome))== -1)
            error("In the file %s the a chromosome value listed (%s) is not valid or is"
               " not present in all simulation files", (const char *) info.filesToRead[sim],
                (const char *) chromosome);
         chr_index = info.chromosomeLookup.Integer(chr_index);
         }
      last_chromosome = chromosome;

      curr_position.printf("%s:%s", (const char *) tokens[4], (const char *) chromosome);

      if (curr_position != last_position)
         if ((pos_index = info.positionLookup.Find(curr_position)) == -1)
            {
            if (!info.gridOn || info.skippedLookup.Find(curr_position) == -1)
               error("In the file %s the position value listed (%s) is"
                     " not found in all simulation files",
                     (const char *) info.filesToRead[sim], (const char *) curr_position);
            continue;
            }

      curr_family_name = tokens[1];
      if (curr_family_name != last_family_name)
         if ((fam_index = info.familyLookup.Find(curr_family_name)) == -1)
            error("In the file %s the family listed (%s) is not present in all simulation files\n",
                 (const char *) info.filesToRead[sim], (const char *) curr_family_name);

      last_family_name == curr_family_name;

      if (info.chromosomeNames[chr_index] != chromosome)
         error("In the file %s the chromosome value listed for position %s on chromosome %s"
            " doesn't correspond to that listed in file %s\n",
             (const char *) info.filesToRead[sim], (const char *) tokens[4],
             (const char *) chromosome, (const char *) info.filesToRead[0]);
      last_position = curr_position;

      test_value = tokens[5];
      zscore = strtod(test_value, &flag);
      if (*flag)
         error("In the file %s, the zscore listed (%s) for family %s "
            "at position %s is not a valid number\n",
             (const char *) info.filesToRead[sim], (const char *) tokens[5],
             (const char *) curr_family_name, (const char *) tokens[4]);

      test_value = tokens[9];
      min = strtod(test_value, &flag);
      if (*flag)
         error("In the file %s, the min value listed (%s) for family %s "
            "at position %s is not a valid number\n",
            (const char *) info.filesToRead[sim], (const char *) tokens[9],
            (const char *) curr_family_name, (const char *) tokens[4]);

      test_value = tokens[10];
      max = strtod(test_value, &flag);
      if (*flag)
         error("In the file %s, the max value listed (%s) for family %s "
            "at position %s is not a valid number\n",
            (const char *) info.filesToRead[sim], (const char *) tokens[10],
            (const char *) curr_family_name, (const char *) tokens[4] );

      affectionScores[aff_index][sim][fam_index][pos_index] = zscore;
      affectionMins[aff_index][sim][fam_index][chr_index] = min;
      affectionMaxs[aff_index][sim][fam_index][chr_index] = max;
   }
   fclose(curr_file);
   }

void PseudoZScorePool::ReadRawZScoreData()
   {
   int n_files = info.filesToRead.Length();

   for (int sim = 0; sim < n_files; sim++)
      {
      printf("Reading information from file %s\n", (const char *) info.filesToRead[sim]);
      ReadRawZScoreFile(sim);
      }
   info.skippedLookup.Clear();
   }

void PseudoZScorePool::ReadRawZScoreFile(int sim)
   {
   FILE * curr_file = fopen((const char *) info.filesToRead[sim], "rb");

   if (!curr_file)
      error("PseudoSimulator::Couldn't find simulation result file %s",
               (const char *) info.filesToRead[sim]);

   String       buffer;
   StringArray  tokens, position_line;

   bool     skip = false;

   int      aff_index = -1, chr_index = -1;
   String   curr_chromosome;

   IntArray position_indices;

   while (!feof(curr_file))
      {
      buffer.ReadLine(curr_file);

      tokens.Clear();
      if (tokens.AddTokens(buffer, WHITESPACE) < 1) continue;

      if (skip && tokens[0] != "ANALYSIS" && tokens[0] != "CHROMOSOME")
         continue;

      if (tokens[0] == "CHROMOSOME")
         {
         chr_index = ReadZChromosomeLine(tokens, curr_chromosome, sim);
         skip = false;
         continue;
         }

      if (tokens[0] == "POSITIONS")
         {
         position_line = tokens;
         position_indices.Dimension(position_line.Length() - 1);
         continue;
         }

      if (tokens[0] == "ANALYSIS")
         {
         aff_index = ReadZAffectionLine(tokens);
         if (skip = (aff_index == -1))
            continue;

         DeterminePositionIndices(position_line, position_indices, curr_chromosome, sim);
         continue;
         }

      if (tokens[0] == "FAMILY")
         ReadZFamilyLine(tokens, position_line, position_indices, aff_index, chr_index, sim);
      }

      fclose(curr_file);
   }

int PseudoZScorePool::ReadZChromosomeLine(StringArray & tokens, String & curr_chromosome, int sim)
   {
   int chr_index;
   curr_chromosome = tokens[1];
   if ((chr_index = info.chromosomeLookup.Find(curr_chromosome)) == -1)
      error("In the file %s the chromosome value listed (%s) is not valid or is"
            " not present in all simulation files", (const char *) info.filesToRead[sim],
            (const char *) curr_chromosome);

   return info.chromosomeLookup.Integer(chr_index);
   }

int PseudoZScorePool::ReadZAffectionLine(StringArray & tokens)
   {
   int spot = tokens[1].FindChar(':');

   String curr_affection = tokens[1].SubStr(0, spot);
   String curr_score = tokens[1].SubStr(spot + 1);

   int score_type = info.GetScoreType((const char *) curr_score);
   info.SelectAffectionName(curr_affection, score_type);

   return info.affectionLookup.Find(info.workingAffectionName);
   }

void PseudoZScorePool::DeterminePositionIndices(StringArray & position_line,
 IntArray & position_indices, String & curr_chromosome, int sim)
   {
   position_indices.Set(-1);

   String curr_position;
   int    pos_index;
   int    n_positions = position_indices.Length();
   String chr_tag = ":" + curr_chromosome;

   for (int pos = 0; pos < n_positions; pos++)
      {
      curr_position = position_line[pos + 1] + chr_tag;
      if ((pos_index = info.positionLookup.Find(curr_position)) == -1)
          if (!info.gridOn ||  info.skippedLookup.Find(curr_position) == -1)
              error("In the file %s the position value listed (%s) is not found in all"
              " simulation files", (const char *) info.filesToRead[sim],
              (const char *) position_line[pos + 1]);

      position_indices[pos] = pos_index;
      }
   }


void PseudoZScorePool::ReadZFamilyLine(StringArray & tokens, StringArray & position_line,
 IntArray & position_indices, int aff_index, int chr_index, int sim)
   {
   const char * test_value;
   char *       flag = NULL;
   double       min, max, zscore;

   int fam_index;
   int n_tokens = tokens.Length();

   if ((fam_index = info.familyLookup.Find(tokens[1])) == -1)
            error("In the file %s the family listed (%s) is not present in all simulation files\n",
            (const char *) info.filesToRead[sim], (const char *) tokens[1]);

   //fam_index = info.familyLookup.Integer(fam_index);

   test_value = tokens[3];
   min = strtod(test_value, &flag);
   if (*flag)
        error("In the file %s, the min value listed (%s) for family %s "
               "is not a valid number\n",
               (const char *) info.filesToRead[sim], (const char *) tokens[3],
               (const char *) tokens[1]);

   test_value = tokens[5];
   max = strtod(test_value, &flag);
   if (*flag)
      error("In the file %s, the max value listed (%s) for family %s "
           "is not a valid number\n",
            (const char *) info.filesToRead[sim], (const char *) tokens[5],
            (const char *) tokens[1]);

   affectionMins[aff_index][sim][fam_index][chr_index] = min;
   affectionMaxs[aff_index][sim][fam_index][chr_index] = max;

   if (n_tokens - 7 != position_indices.Length())
      error("PseudoSimulator::Number of z-scores listed does not match number"
            "of positions listed for file %s", (const char *) info.filesToRead[sim]);

   for (int tok = 7; tok < n_tokens; tok++)
      {
      if (position_indices[tok - 7] != -1)
         {
         test_value = tokens[tok];
         zscore = strtod(test_value, &flag);
         if (*flag)
             error("In the file %s, the family lod value listed (%s) for family %s "
                        "at position %s is not a valid number\n",
                        (const char *) info.filesToRead[sim], (const char *) tokens[tok],
                        (const char *) tokens[1],
                        (const char *) position_line[position_indices[tok - 7]] );

          affectionScores[aff_index][sim][fam_index][position_indices[tok - 7]] = zscore;
          }
      }
   }

void PseudoZScorePool::AllocateInputArrays()
   {
   affectionMaxs = new Matrix *[info.nAffections];
   affectionMins = new Matrix *[info.nAffections];
   affectionScores = new Matrix *[info.nAffections];

   for (int aff = 0; aff < info.nAffections; aff++)
      {
      affectionScores[aff] = new Matrix[nPools];
      affectionMins[aff] = new Matrix[nPools];
      affectionMaxs[aff] = new Matrix[nPools];

      for (int sim = 0; sim < nPools; sim++)
         {
         affectionMaxs[aff][sim].Dimension(info.nFamilies, info.nChromosomes);
         affectionMaxs[aff][sim].Zero();

         affectionMins[aff][sim].Dimension(info.nFamilies, info.nChromosomes);
         affectionMins[aff][sim].Zero();

         affectionScores[aff][sim].Dimension(info.nFamilies, info.nAllPositions);
         affectionScores[aff][sim].Zero();
         }
      }
   }

void PseudoZScorePool::FreeInputArrays()
   {
   if (affectionScores != NULL)
      {
      for (int aff = 0; aff < info.nAffections; aff++)
         if (affectionScores[aff]) delete [] affectionScores[aff];
      delete [] affectionScores;
      }
   affectionScores = NULL;

   if (affectionMins != NULL)
      {
      for (int aff = 0; aff < info.nAffections; aff++)
         if (affectionMins[aff]) delete [] affectionMins[aff];
      delete [] affectionMins;
      }
   affectionMins = NULL;

   if (affectionMaxs != NULL)
      {
      for (int aff = 0; aff < info.nAffections; aff++)
         if (affectionMaxs[aff]) delete [] affectionMaxs[aff];
      delete [] affectionMaxs;
      }
   affectionMaxs = NULL;
   }

double PseudoZScorePool::EstimateArrayMemory()
   {
   MemoryEstimator est;

   // affectionScores
   double mem_needed = info.nAffections * sizeof(Matrix *);
   mem_needed += info.nAffections * nPools * est.MemoryForMatrix(info.nFamilies, info.nAllPositions);

      // affectionMins, affectionMaxs
   mem_needed += 2 * info.nAffections * sizeof(Matrix *);
   mem_needed += 2 * info.nAffections * nPools * est.MemoryForMatrix(info.nFamilies, info.nChromosomes);

   return mem_needed;
   }


 
 
 
